;       Last change    August 1993
(in-package epilib)
;;; The net unification package. Provides a general purpose unification
;;; routine.

(tweakable '*fix-unify* t "testing parameter only" 'boolean)


;;; During unification, a tree-like data structure is used to keep track
;;; of which argument is currently subsituting for other arguments.  The
;;; root of the tree (or the 'name' of the set which the tree represents)
;;; substitutes all other arguments in the tree (or members in the set).
;;; Access to this structure consists entirely of union and find operations.
;;; Using a union-finf algorithm with near linear time complexity allows the
;;; unification algorithm also to be near linear.  The algorithm used is
;;; described in "The Design and Analysis of Computer Algorithms", by
;;; Aho et. al.
;;;

;;; Union find constants

(defconstant uf-table-size 100)


;;; The same union-find table used throughout--it's 'zeroed' after every
;;; unification attempt.
;;;
(defconstant uf-elts (make-array uf-table-size :initial-element nil))
(defvar *uf-used-indices* nil)
(defvar *uf-functions* nil)


;;; Union-find table element.  
;;;
(defstruct uf-elt
	arg		;arg which corresponds to entry
	clause-num	;either 1 or 2
	keep-quant      ; t if quantifier should be kept after sub
	(num-mem 1)	;number of elements in the set of which this element
			; is a root. If 0, this element is not a root.
	subs		;if this element is a root, then this  points to the 
			; set member which is the current substitution for all
			; other elements in the set; otherwise it points to
			; the father of this element
)

(defmacro ref-elt (index) `(aref uf-elts ,index))

;;; Returns the root element of the set in which the given arg is a
;;; member.  A new set is created if necessary.
 
(defun uf-get (arg clause-num &aux elt-index elt)
    (if (setq elt (ref-elt (setq elt-index (uf-hash arg clause-num))))
	(uf-find-root elt)
	(progn
	    (push elt-index *uf-used-indices*)
	    (setf (ref-elt elt-index) 
		  (setq elt (make-uf-elt :arg arg :clause-num clause-num)))
	    (setf (uf-elt-subs elt) elt)
	    (if (equal (function-p arg) 'function)
		(push elt *uf-functions*))
	    elt))
  )


(defun uf-hash (arg clause-num &aux conc-type probe elt)
  (setq probe (sxhash arg))

  (setq conc-type (term-type arg))

    (if (member conc-type '(variable function))
	(setq probe (* probe clause-num 123))
	(setq clause-num nil))

    (setq probe (mod probe uf-table-size))
    
    ;;; Linear probe until we find the arg or an empty slot

    (do nil
	((or (not (setq elt (ref-elt probe)))
	     (and (eq (uf-elt-arg elt) arg)
		  (or (not clause-num)
		      (eql (uf-elt-clause-num elt) clause-num))))
	  probe)
	(setq probe (mod (+ 1 probe) uf-table-size)))
  )

;;; Finds the root element of a given element.  Paths compression is done
;;; after the search.

(defun uf-find-root (elt &aux (path nil))
    (do nil
	((> (uf-elt-num-mem elt) 0))
	(push elt path)
	(setq elt (uf-elt-subs elt)))

    (mapc #'(lambda (mem)
	          (setf (uf-elt-subs mem) elt))
	  path)
    
    elt)

;;; Performs the union of two sets.  The subsititution element of the
;;; union will be the current substitution element of the first set.
;;; The union is done by adding the smaller of the two sets as a new
;;; branch off the root of the larger.

(defun uf-union (root-2 root-1 &optional fixed keep-quant)
    (if (or fixed (>= (uf-elt-num-mem root-1) (uf-elt-num-mem root-2)))
	(progn
	    (incf (uf-elt-num-mem root-1) (uf-elt-num-mem root-2))
	    (setf (uf-elt-num-mem root-2) 0)
	    (setf (uf-elt-keep-quant root-1)  keep-quant)
	    (setf (uf-elt-subs root-2) root-1))
	(progn
	    (setf (uf-elt-subs root-2) (uf-elt-subs root-1))
	    (incf (uf-elt-num-mem root-2) (uf-elt-num-mem root-1))
	    (setf (uf-elt-num-mem root-1) 0)
	    (setf (uf-elt-keep-quant root-2)  keep-quant)
	    (setf (uf-elt-subs root-1) root-2)))
    t)


(defun compare-cleanup ()
   ;; Make sure everything's cleaned up
   (when *uf-used-indices*
       (mapc #'(lambda (index) (setf (ref-elt index) nil))
	     *uf-used-indices*)
       (setq *uf-used-indices* (setq *uf-functions* nil)))

    (do ((elt *uf-functions* (cdr elt)))
	((or (null elt) ))
	(setf (ref-elt elt) nil))

    (setq *uf-used-indices* (setq *uf-functions* nil))
  )


(tweakable '*unify-sorts* t 
  "Indicates whether sorts should be checked during unification"
  'boolean)

(defun compatible-sorts (sort-a sort-b &optional (null-doesnt-match nil))
  (cond ((equal sort-a sort-b) t)
	((or (null sort-a) (null sort-b))  ; no sort matches anything?
	 (not null-doesnt-match))
	(t (member (hier-rel sort-a sort-b)
	      '(subsumes equivalent)))
	)
  )

(defun compare-variables (arg-a arg-b &optional prefer keep-quant exist-match
				&aux root-a root-b)
  (or (eq (setq root-a (uf-get arg-a 1))
	  (setq root-b (uf-get arg-b 2)))
      (unify-var root-a root-b prefer keep-quant exist-match))
  )

(defun unify-var (root-a root-b &optional prefer keep-quant exist-match
		                &aux (subs-a (uf-elt-subs root-a))
				     (subs-b (uf-elt-subs root-b))
				     (arg-a (uf-elt-arg subs-a))
				     (arg-b (uf-elt-arg subs-b))
				     (type-a (entity-type arg-a))
				     (type-b (entity-type arg-b))
				     (sort-a (find-sort arg-a))
				     (sort-b (find-sort arg-b)))

  (declare (ignore exist-match)) ;; this is temporary--see note marked CHANGE below
  (if (and (epi-variable arg-a) (getp arg-a 'MATCHABLE))
      (if (and (epi-variable arg-b) (getp arg-b 'MATCHABLE))
	  (if *unify-sorts*
	      (if prefer
		  (if (equal prefer 1)
		      (if (compatible-sorts sort-a sort-b)
			  (uf-union root-b root-a t keep-quant))
		      (if (compatible-sorts sort-b sort-a)
			  (uf-union root-a root-b t keep-quant)))
		  (if (compatible-sorts sort-b sort-a) 
		      (uf-union root-a root-b t keep-quant)
		      (if (compatible-sorts sort-a sort-b) 
			  (uf-union root-b root-a t keep-quant))) )
	      (if prefer
		  (if (equal prefer 1) (uf-union root-a root-b t keep-quant)
		      (uf-union root-b root-a t keep-quant))
		  (uf-union root-a root-b nil keep-quant)))
	  ; arg b is either a constant or a non-matchable variable
	  (if (if (and (not (equal type-a type-b)) (member sort-a '(wff term quantifier operator function meta-entity pred)) type-b) (compatible-sorts sort-a type-b) t)
	      (if (if *unify-sorts* (compatible-sorts sort-a sort-b) t)
		  (uf-union root-a root-b t keep-quant))
	      ))
      ; arg-a is either a constant or a non-matchable variable
      (if (and (epi-variable arg-b) (getp arg-b 'MATCHABLE))
	  (if (if (and (not (equal type-a type-b)) (member sort-b '(wff term quantifier operator function emta-entity pred)) type-a) (compatible-sorts sort-b type-a) t)
	      (if (if *unify-sorts* (compatible-sorts sort-b sort-a) t)
		  (uf-union root-b root-a t keep-quant))
	      )
	  ;neither is a matchable variable - either lambdas or E's
	  ; if constants, try to unify still, if unmatchable vars, dont
	  ;;;; NOTE lambda vars are unmatchable, but should still be
	  ;;;; unifiable with each other 
	  ;;;CHANGE - allow non-matchable identical variables to unify
	  ;; needed for comparisons of complex episodic junk - fix!
	  (if (and (epi-variable arg-a) (epi-variable arg-b)
		   (or ;exist-match
		       (and (getp arg-a 'LAMBDA) (getp arg-b 'LAMBDA))
		       (and (equal arg-a arg-b)
			    (not (getp arg-a 'LAMBDA)) 
			    (not (getp arg-b 'LAMBDA)))))
	      (uf-union root-a root-b nil keep-quant)
	      nil)
	  )
      )
  )

(defun print-trace-subs (subs &optional args)
    (if (not subs)
	(print-it "nothing ")
	(mapc #'(lambda (sub) (print-it (cadr sub) "/" (car sub) " " )) subs)
	)
  (when args (print-line "         in " args))
  )

(defun print-subs (sublist)
      (mapcar #'(lambda (a) (list (print-info (car a)) "by" (print-info (cadr a)) (caddr a))) sublist)
    )

