;;; The EpiShell interp package--implements the EpiShell interpretor.
(in-package user)
(eval-when (compile load eval)   
(provide 'esh-interp)
(require 'esh-io "epilog:eshell;io")
(require 'esh-read "epilog:eshell;read")
(require 'esh-cmdhash "epilog:eshell;cmdhash")
(defpackage esh-interp 
  (:use lisp #+lucid lucid-common-lisp #-lucid common-lisp #+excl excl
	esh-io esh-read esh-cmdhash)
  (:export
          ;; Parser variables

          *esh-take-char* *esh-put-char* *esh-pipe-char*
          *esh-comment-char* *esh-include-char*

	  esh epi-shell *esh-help-path* *esh-command-list*
	 ))

(unless (find-package 'esh-interp) (make-package 'esh-interp)))
(in-package esh-interp)

;(defvar *esh-take-char* #\<
;        "Input redirection character for EpiShell" )
;
;(defvar *esh-put-char* #\>
;        "Output redirection character for EpiShell" )
;
;(defvar *esh-pipe-char* #\| 
;        "Process pipe character for EpiShell" )
;
;(defvar *esh-comment-char* #\;
;        "EpiShell comment delimiter" )
;
;(defvar *esh-include-char* #\@
;        "File include character for EpiShell" )

(defvar *esh-help-path* 
#+lucid  "epilog:help;.hlp"
#-lucid  "epilog:help;<help>.hlp"
	  "Pathname for EpiShell help text files" )

(defvar *esh-command-list* nil
  "Command list of the current EpiShell, to be used when starting a new one. In reverse order.")

(defun esh-version () "1.0" )

(defvar *esh-open-files* nil
	"Names of all the files currently open by EpiShell(s)" )

;;; Structure used to describe a shell environment.
;;
(defstruct esh-env
    (cmd-hash-table nil)	;command hash table
    (prompt "E> ")		;prompt string
    (input-stack nil)		;stack of files which are being read for shelln
				; input
    (default-fn nil)            ;function to invoke when given a simple list
    (default-prompt "D> ")      ;prompt to use when default is active
    (echo-flag nil)		;true if input files should be echoed
    (trace-stream nil)		;stream being used for trace output
    (output-stream nil)		;stream being used for standard output
)

;;; esh is the easy way to start up a EpiShell.
;;;
(defun esh (&optional (input-file-or-stream-or-cmd *standard-input*)
                      (profile-file 'epishell.pro) &aux fullname ) 
         "Start up a new EpiShell"
    (epi-shell input-file-or-stream-or-cmd *standard-output* nil nil nil
	       (if (and (setq fullname (get-filename profile-file))
			(probe-file fullname))
                    profile-file))
)


;;; The EpiShell itself. 
;;;
(defun epi-shell (&optional input-stream output-stream 
			    error-stream query-stream 
			    trace-stream profile-file 
		  &aux  (save-standard-input *standard-input*)
			(save-standard-output *standard-output*)
			(save-error-output *error-output*)
			(save-query-io *query-io*)
			(save-trace-output *esh-trace-output*)
			(save-command-list *esh-command-list*) 
  		        (save-open-files *esh-open-files*)
                        (save-include-char *esh-include-char*)
                        (save-take-char *esh-take-char*)
                        (save-put-char *esh-put-char*)
                        (save-comment-char *esh-comment-char*)
			(env (make-esh-env))
			(created-input-stream nil)
		)
    "Start up a new EpiShell"
  
(cond ((and output-stream 
            (not (and (streamp output-stream) 
                                            (output-stream-p output-stream))))
	  (err-msg "'" output-stream 
		      "' is an invalid output stream for a EpiShell."))
      ((and error-stream
	    (not (and (streamp error-stream) (output-stream-p error-stream))))
	  (err-msg "'" error-stream 
		      "' is an invalid error output stream for a EpiShell."))
      ((and query-stream 
	    (not (and (streamp query-stream) (input-stream-p query-stream)
					     (output-stream-p query-stream))))
	  (err-msg "'" query-stream 
		      "' is an invalid query stream for a EpiShell."))
      ((and trace-stream
	    (not (and (streamp trace-stream) (output-stream-p trace-stream))))
	  (err-msg "'" trace-stream 
		      "' is an invalid trace output stream for a EpiShell."))
(t
  (unwind-protect (prog nil

    (if (and input-stream (not (streamp input-stream)))
    	(setq input-stream
                (setq created-input-stream 
	            (make-string-input-stream
			(if (stringp input-stream)
			    (concatenate 'string input-stream 
					 (string #\newline))
			    (concatenate 'string (string *esh-take-char*)
    				         (princ-to-string input-stream)
					 (string #\newline)))))))

    (setq *query-io*
	(or query-stream 
            (if (and input-stream (output-stream-p input-stream))
	        input-stream
		(if (output-stream-p *standard-input*)
		    *standard-input*
	            *terminal-io* ))))

    (if input-stream	
	(setq *standard-input* input-stream))

    (if output-stream
	(setq *standard-output* output-stream))
    (setf (esh-env-output-stream env) *standard-output*)

    (setq *error-output*
	(or error-stream *standard-output*))

    (setq *esh-trace-output*
	(or trace-stream (esh-env-output-stream env)))

    (setf (esh-env-input-stack env) `( ,`( ,*standard-input* )))
    (if (and profile-file 
  	     (setq profile-file (shell-open profile-file :input)))
	(push `(,profile-file) (esh-env-input-stack env)))

				
    ;; Save the hash table for faster entry next time
    ;;
    (if (not (get '*esh-command-list* '*esh-command-table*))
        (setf (get '*esh-command-list* '*esh-command-table*)
              (make-cmd-hash-table (reverse *esh-command-list*))))
    (setf (esh-env-cmd-hash-table env)
          (get '*esh-command-list* '*esh-command-table*))

    (do nil ((not (esh-env-input-stack env)))
	(setq *standard-input* (caar (esh-env-input-stack env)))

	(if (and (output-stream-p *standard-input*)
			(not (cdr (esh-env-input-stack env))))
                (display-prompt env))
	(do ((command (read-next *standard-input* *esh-open-files*)))
            ((and (listp command) (eq (car command) esh-eof)))

	    (if (eq command esh-read-error)
		(setq command nil)
                (setq command (do-command env command)))

            (when (not command)
		(if (and (output-stream-p *standard-input*)
				(not (cdr (esh-env-input-stack env))))
                    (display-prompt env))
	        (setq command 
		       (read-next *standard-input* *esh-open-files*)))
        )

	(let ((last-input (pop (esh-env-input-stack env))))
	    (if (esh-env-input-stack env) ; don't close the one we're given
                                          ; i.e., it's still on the stack
	        (if (cdr last-input)
		    (progn
	               (close (car last-input))
		       (shell-close (cadr last-input)))
	            (shell-close (car last-input)))))
    ))
    (setq *standard-input* save-standard-input)
    (setq *standard-output* save-standard-output)
    (setq *error-output* save-error-output)
    (setq *query-io* save-query-io)
    (setq *esh-trace-output* save-trace-output)
    ;; Can't save the table if it's been changed
    (when (not (equal *esh-command-list* save-command-list))
          (remprop '*esh-command-list* '*esh-command-table*)
          (setq *esh-command-list* save-command-list))
    (setq *esh-open-files* save-open-files)

    (setq *esh-include-char* save-include-char)
    (setq *esh-take-char* save-take-char)
    (setq *esh-put-char* save-put-char)
    (setq *esh-comment-char* save-comment-char)
   
    ;; Reset read syntax to the default

    (reset-readtables)

    ;; Try to close off all the files we've opened

    (do ((last-input (pop (esh-env-input-stack env))
		     (pop (esh-env-input-stack env))))
	((not (esh-env-input-stack env)))
	(if (esh-env-input-stack env)     ; don't close the one we're given
                                          ; i.e., it's still on the stack
	    (if (cdr last-input)
		(progn
	           (close (car last-input))
		   (shell-close (cadr last-input)))
	        (shell-close (car last-input)))))

    (when (esh-env-trace-stream env)
          (print-msg "Closed trace file " 
				(get-filename (esh-env-trace-stream env)))
	  (shell-close (esh-env-trace-stream env)))

    (if created-input-stream (close created-input-stream))
))))


(defun display-prompt (env)
    (fresh-line)
    (princ (if (esh-env-default-fn env)
               (esh-env-default-prompt env)
               (esh-env-prompt env)) *standard-input*)
)

;;; eshell-echo-stream creates a new echo stream, given the shell environment
;;; and an input stream.
;;; 
(defun eshell-echo-stream (env input-stream)
    (make-echo-stream input-stream (esh-env-output-stream env))
)


;;; do-command is called to parse and execute a command line
;;; If the line is of the form <filename, then all that happens is that
;;; the shell's input is redirected; otherwise, it's a 'real' command.
;;;
(defun do-command (env command &aux cmd-tokens take-stream)
    (setq command (parse-command command))
    (if (not (setq cmd-tokens (cdr (assoc 'cmd command)))) ; no command on line
	(when (and (cdr (assoc 'take command))
		   (setq take-stream 
				(shell-open (cdr (assoc 'take command))
					      :input)))
	      (setq take-stream (list take-stream))
	      (if (esh-env-echo-flag env)
			(setq take-stream 
				(cons (eshell-echo-stream env (car take-stream) )
				       take-stream)))
	      (setq *standard-input* 
			(caar (push take-stream (esh-env-input-stack env))))
	      nil)
        (do-redirection env (cdr (assoc 'take command))
                            (cdr (assoc 'put command))
                            (cdr (assoc 'append command))
                            (cdr (assoc 'pipe command))
                            (car cmd-tokens) (cdr cmd-tokens)))
   (cdr (assoc 'next-cmd command))
)


;;; do-redirection does input file redirection as 
;;; indicated on a shell command line and executes the command.
;;;
(defun do-redirection (env take put app pipe cmd cmd-args)
    (if take 
	(if (setq take (check-filename take *esh-open-files*))
	    (let ((save-stream *standard-input*))
	        (unwind-protect
		    (with-open-file (take-stream take :if-does-not-exist nil)
		        (cond	
			    (take-stream
		                (mark-as-open take)
			        (setq *standard-input* take-stream)
			        (do-output-redirection env put app pipe cmd 
                                                                     cmd-args))
			    (t (file-open-msg *standard-input* take))))
		    (setq *standard-input* save-stream)
		    (mark-as-closed take))))
	(do-output-redirection env put app pipe cmd cmd-args))
)


;;; do-output-redirection does output file redirection as 
;;; indicated on a shell command line and executes the command.
;;;
;;; this function uses SYSTEM DEPENDENT code to implement pipes to
;;; a Unix process
;;;
(defun do-output-redirection (env put app pipe cmd cmd-args)
    (if pipe
        (let ((save-stream *standard-output*)  tmp-stream
              (tmp-file (check-filename 
                             (concatenate 'string  "/tmp/pipe." 
                                     (string-downcase (string (gensym))))
                             *esh-open-files*)))
            (unwind-protect
                (progn 
                    (setq tmp-stream (open tmp-file :direction :output
			               :if-does-not-exist  :create
			               :if-exists :new-version))
	            (cond (tmp-stream
		          (mark-as-open tmp-file)
			  (setq *standard-output* tmp-stream)
		          (execute-cmd env cmd cmd-args)
                          (setq *standard-output* save-stream)
                          (close tmp-stream)
                          #+lucid (lucid::run-unix-program "/bin/csh" 
                                :arguments `("-c" ,(concat-args "cat" 
                                                           tmp-file
                                                           "|"  pipe))
                                :input  (if (get-filename *standard-input*)
                                            *standard-input*)
                                :output (if (get-filename *standard-output*)
                                            *standard-output*)
	                        :error-output (if (get-filename *error-output*)
                                                  *error-output*))
                          #+allegro (excl:run-shell-command 
                                `("/bin/csh -c" ,(concat-args "cat" 
                                                           tmp-file
                                                           "|"  pipe))
                                :input  (if (get-filename *standard-input*)
                                            *standard-input*)
                                :output (if (get-filename *standard-output*)
                                            *standard-output*)
	                        :error-output (if (get-filename *error-output*)
                                                  *error-output*))
		          (fresh-line))
		      (t (file-open-msg *standard-input* tmp-file))))
            (setq *standard-output* save-stream)
            (if tmp-stream (delete-file tmp-file))
	    (mark-as-closed tmp-file)) )
        
        (if (setq put (or app put))
	    (if (setq put (check-filename put *esh-open-files*))
	        (let ((save-stream *standard-output*))
	            (unwind-protect
		        (with-open-file (put-stream put :direction :output
			      :if-does-not-exist  :create
			      :if-exists (if app :append :new-version) )
		            (cond	
			        (put-stream
		                    (mark-as-open put)
			            (setq *standard-output* put-stream)
				    (execute-cmd env cmd cmd-args)
                                    (setq *standard-output* save-stream)
				    (fresh-line))
			        (t (file-open-msg *standard-input* put))))
		        (setq *standard-output* save-stream)
		        (mark-as-closed put))))
	    (execute-cmd env cmd cmd-args)))
)

;;; concat-args concatenates its arguments into a single string, with
;;; each argument separated by a blank
;;;
(defun concat-args (&rest args)
    (reduce #'(lambda (str arg)
                   (concatenate 'string str
                           (cond ((not arg) "")
                                 ((stringp arg) arg)
                                 ((listp arg) (apply #'concat-args arg))
                                 (t (string-downcase (princ-to-string arg))))
                           " "))
            args :initial-value "")
)


;;; execute-command does the actual command, after all input/output
;;; redirection has been done.
;;; If the command is a list, then if a default function has been set, 
;;; the list is given to that function; otherwise, it is evaluated as 
;;; a Lisp form.
;;; Else, if it's a shell command, then execute as a command. (Note that
;;; the three `builtin' commands are handled specially, as they use or 
;;; modify the environment.)  Otherwise, if it is the only token on
;;; the line, print its value.  If neither of these, execute as a
;;; Lisp form (without evaluating the arguments).
;;;
;;; In all cases, if the command returns a result, it is printed
;;; onto the standard output
;;;
(defun execute-cmd (env cmd cmd-args &aux result shell-cmd (display-nil nil))
    (when (or (setq result
        (if (not (atom cmd))
            (esh-eval (let ((cmd-function (esh-env-default-fn env)))
                       (if cmd-function 
                           `(,cmd-function ,`(lisp::quote ,cmd))
                           (progn (setq display-nil t) cmd))))
            (if (setq shell-cmd 
                      (get-cmd-hash cmd (esh-env-cmd-hash-table env)))
                (cond
	            ((member shell-cmd '(esh-set esh-show))
	                (setshow env shell-cmd cmd-args))
	            ((eq shell-cmd 'esh-help)
	                (esh-help env cmd-args))
                    (t (esh-eval (cons shell-cmd (quote-args cmd-args)))))
                (esh-eval (if cmd-args (cons cmd (quote-args cmd-args)) 
                                       (progn (setq display-nil t) cmd))))))
           display-nil)
        (fresh-line) (prin1 result) (terpri))
)

(defun quote-args (args)
    (mapcar #'(lambda (arg) `(lisp::quote ,arg)) args))


;;; esh-eval uses SYSTEM DEPENDENT code to allow an abort back to
;;; the shell when evalating an expression.
;;;
(defun esh-eval (expr)
  #+lucid (lucid::with-simple-restart (nil "Abort to EpiShell" nil) (eval expr))
  #+allegro (eval expr)
   )


;;; parse-command parses a shell command line, breaking it up into
;;; the command itself and file redirection or pipe directives. 
;;; If the first non-directive token on the line is a list, then 
;;; other non-directive tokens are assumed to belong to the next
;;; command.
;;;
(defun parse-command (command &aux (next-cmd nil) (take nil) (put nil) 
				      (append nil) (pipe nil) (cmd nil))
    (do ((token (car command) (car command)))
	((or next-cmd (not command)))
	
	(cond ((eql token *esh-pipe-char*)
		    (setq pipe (cdr command))
                    (setq command nil))
              ((eql token *esh-take-char*)
		    (setq command (cdr command))
		    (unless (or (not command) (not (atom (car command)))
			        (eql (car command)  *esh-put-char*)
                                (eql (car command)  *esh-pipe-char*))
			(setq take (car command))
			(setq command (cdr command))))
	      ((eql token *esh-put-char*)
		    (setq command (cdr command))
		    (unless (or (not command) (not (atom (car command)))
				(eql (car command) *esh-take-char*)
                                (eql (car command)  *esh-pipe-char*))
			(cond ((eql (car command) *esh-put-char*)
				    (setq command (cdr command))
				    (unless (or (not command) 
						(not (atom (car command))))
					(setq append (car command))))
			      (t
				    (setq put (car command))))
			(setq command (cdr command))))
	      (t (if (and cmd (not (cdr cmd)) (not (atom (car cmd))))
                     (setq next-cmd command)
                     (progn 
                          (setq cmd (cons token cmd))
		          (setq command (cdr command))))))
    )
    `( ,`(take . ,take) ,`(put . ,put) ,`(append . ,append) ,`(pipe . ,pipe)
		,`(cmd . ,(nreverse cmd)) ,`(next-cmd . ,next-cmd))
)


;;; shell-open opens file for the EpiShell.  If successful, it marks the
;;; file as open, so that it cannot be opened again, unless it is first
;;; closed.
;;;
(defun shell-open (filename access-type &aux stream)
    (when (setq filename (check-filename filename *esh-open-files*))
	(cond ((setq stream (open-file filename access-type))
		   (mark-as-open filename)
		   stream)
	      (t   (file-open-msg *standard-input* filename) ) ) )
)

(defun mark-as-open (filename)	
    (setq *esh-open-files* (cons filename
					*esh-open-files*))
    filename)

(defun shell-close (stream)
    (mark-as-closed (get-filename stream))
    (close stream)
)

(defun mark-as-closed (filename)
    (setq *esh-open-files* (delete filename *esh-open-files* 
						 :test #'equal))
    filename)


    

;;; The 'special' shell commands

;;; The help command.
;;;
(defun esh-help (env args) "EpiShell help command"
    (if (not args)
	(err-msg "Use 'show commands' or 'apropos -string-' to get a list of commands "
		    nil "recognized by this shell." )
	(let ((cmd (get-cmd-hash (car args) (esh-env-cmd-hash-table env))))
	    (if (not cmd)
		(err-msg  (car args) " is not a command or command "
			    "abbreviation in this shell.")
		(let ((fname (merge-pathnames (case-filename cmd) 
				              *esh-help-path*)))
		    (with-open-file (file-stream fname 
						:if-does-not-exist nil)
                        (cond
                            (file-stream
   				 (do ((line (read-line file-stream nil esh-eof)
	      			            (read-line file-stream nil 
                                                                     esh-eof)))
       				     ((eq line esh-eof) nil)
				     (write-line line)))
                            (t (file-open-msg *standard-input* fname))))))
	))
)


;;;; Set and show commands--fairly straighforward code.

(defvar *esh-set-options* (make-cmd-hash-table 
    '((commands) (comment-char)  
      (default-function def-function) (default-prompt def-prompt) 
      (default-syntax def-syntax) (noecho) (echo)
      (include-char)  (prompt) (pipe-char) (put-char) 
      (symbol-syntax)
      (take-char) (trace-output) (version) ) ) 
	"Options for EpiShell set and show commands" )


(defun setshow (env cmd cmd-args &aux option)
    (cond ((not cmd-args)
	      (err-msg "What do you want to set (or show)?"))
	  ((not (setq option (get-cmd-hash (car cmd-args) 
					      *esh-set-options*)))
	      (err-msg (car cmd-args) 
				" is not an option you can set (or show)."))
	  ((eq cmd 'esh-set)
	      (esh-set env option (cdr cmd-args)))
	  (t  (esh-show env option (cdr cmd-args))))
)

(defun esh-set (env option args) "Set values of EpiShell options"
    (case option
        (comment-char 
            (setq *esh-comment-char* (parse-char args *esh-comment-char*))
            (reset-readtables)) 
        (include-char 
            (setq *esh-include-char* (parse-char args *esh-include-char*))
            (reset-readtables)) 
        (take-char 
            (setq *esh-take-char* (parse-char args *esh-take-char*))
            (reset-readtables)) 
        (pipe-char 
            (setq *esh-pipe-char* (parse-char args *esh-pipe-char*))
            (reset-readtables)) 
        (put-char 
            (setq *esh-put-char* (parse-char args *esh-put-char*))
            (reset-readtables)) 
        (symbol-syntax
            (if args
                (let ((char (parse-char args nil)))
                      (if char (set-symbol-syntax char)))
                (print-msg 
        "You should give the character which is to have stand-alone syntax.")))
        (default-syntax
            (if args
                (let ((char (parse-char args nil)))
                      (if char (set-default-syntax char)))
                (print-msg 
        "You should give the character which is revert to a default syntax.")))
 	(commands (set-command env args))
        (default-function
            (setf (esh-env-default-fn env) (if args (car args) nil)))
	(default-prompt
	    (setf (esh-env-default-prompt env) 
			(princ-to-string (if args (car args) "" ))))
	(echo 
	    (setf (esh-env-echo-flag env) t))
	(noecho
	    (setf (esh-env-echo-flag env) nil))
	(prompt
	    (setf (esh-env-prompt env) 
			(princ-to-string (if args (car args) "" ))))
	(trace-output
	    (cond ((not args)
		      (when (esh-env-trace-stream env)
		          (print-msg "Closed trace file " 
				(get-filename (esh-env-trace-stream env)))
		          (shell-close (esh-env-trace-stream env)))
		      (setf (esh-env-trace-stream env) nil)
		      (setq *esh-trace-output* (esh-env-output-stream env)))
		  (t
		      (when (setq args (shell-open (car args) :output))
			    (print-msg "Directing trace output to "
				       (get-filename args))
			    (setf (esh-env-trace-stream env) args)
			    (setq *esh-trace-output* args)))))

	(otherwise (err-msg option " is not something you can set.")))
    nil
)

(defun parse-char (args default)
    (when args 
        (setq args (car args))
        (if (or (characterp args) 
                (and (or (stringp args) (symbolp args))
                     (= (length (string args)) 1)))
            (character args)
            (progn (err-msg "The argument " args
                                " cannot be coerced to a character.")
                   default)))
)
                    
(defun set-command (env names)
    (if (not names)
	(err-msg 
	 "You should supply a command name to set and possibly some synonyms.")
        (do ((name names (cdr name))
	     (good-names nil))
	    ((not name) (make-cmd-hash-table (list (setq good-names
							(nreverse good-names)))
						(esh-env-cmd-hash-table env))
                        (push good-names *esh-command-list*))
	    (unless (and (car name) (symbolp (car name)))
	        (err-msg (car name) 
			" should be a symbol that could be used as a command.")
	        (return nil))
	    (setq good-names (cons (car name) good-names))))
)
	

(defun esh-show (env option args) "Show values of EpiShell setable options"
    (case option
        (comment-char 
	    (print-msg "The current comment character is "
                        (if *esh-comment-char* *esh-comment-char*
                                               "not set."))) 
        (include-char 
	    (print-msg "The current include character is "
                        (if *esh-include-char* *esh-include-char*
                                               "not set."))) 
        (take-char 
	    (print-msg "The current take (from a file) character is "
                        (if *esh-take-char* *esh-take-char*
                                               "not set."))) 
        (pipe-char 
	    (print-msg 
              "The current pipe (to an operating system command) character is "
                        (if *esh-pipe-char* *esh-pipe-char*
                                               "not set."))) 
        (put-char 
	    (print-msg "The current put (to a file) character is "
                        (if *esh-put-char* *esh-put-char*
                                               "not set."))) 
        ((symbol-syntax default-syntax)
            (if args
                (let ((char (parse-char args nil)))
                     (if char
                         (print-msg "This character "
                                  (if (special-char-p char) "has"
                                                            "does not have")
                                 " special syntax in this shell.")))
                (print-msg 
        "You should supply a character for which the syntax is to be shown.")))
	(commands 
	    (if (car args)
	        (let ((cmd (get-cmd-hash (car args) 
					    (esh-env-cmd-hash-table env))))
		     (if (not cmd)
		         (err-msg "'" (car args) 
			         "' is not a command in this shell." nil
				 "Use 'show command' to get the whole list.")
                         (progn 
			     (print-msg "The command " (car args) 
				 " is bound to the Lisp function "
				 (prin1-to-string cmd) "." )
	                     (when (documentation cmd 'function)
                                    (print-msg "    " 
                                             (documentation cmd 'function))))))
               	(progn
	    	    (print-msg "These are the commands (and their synonyms) "
		    	    "recognized by this shell:" )
	    	    (mapc #'(lambda (cmd) 
                                (prin1 (car cmd))
                                (mapc #'(lambda (syn)
                                              (princ " ")
                                              (princ syn))
				       (cdr cmd))
                                (format *standard-output* "~35T")
                                (print-msg 
                                   (if (documentation (car cmd) 'function)
                                       (documentation (car cmd) 'function))))
		  	    (reverse *esh-command-list*)))))
        (default-function
            (if (esh-env-default-fn env)
                (print-msg "The current default function is "
                          (esh-env-default-fn env))
                (print-msg "A default function has not been set.")))
	(default-prompt
	    (print-msg "The current default prompt string is " 
		(prin1-to-string (esh-env-default-prompt env))))
	((echo noecho)
	    (print-msg "Files included by using `"
                (princ-to-string (if *esh-take-char* *esh-take-char* "<"))
                "filename' will "
		(if (esh-env-echo-flag env) "be" "not be") " echoed."))
	(prompt
	    (print-msg "The current prompt string is " 
		(prin1-to-string (esh-env-prompt env))))
	(trace-output
	    (if (esh-env-trace-stream env)
		(print-msg "Trace output is going into file "
		     (get-filename (esh-env-trace-stream env)))
		(print-msg "Trace output is going to *standard-output*")))
	(version (print-msg "You are using version " (esh-version) 
		         " of the EpiShell." ))
	(otherwise (err-msg option " is not something you can show."))) 
    nil
)


