;       Last change    August 1993
(in-package lib)


;Trace functions

; These routines make it simpler to watch the system operate and to debug
; Tracing may be turned off aor on for individual trace items or for all.

; Useful routines:
;    traceable item description trace-values
;       adds item to the list of things which may be traced
;       and puts description on item
;       trace-values indicates that this is a compound trace item and that
;       these items are the ones to be traced for this item
;    traced-p item
;       returns t if the item is currently being traced, nil otherwise
;    trace-item items
;       starts tracing for each items in items. If no items are given, 
;       starts tracing everything
;    trace-all
;       starts tracing everything
;    untrace-item items
;       stops tracing for each item in items. If no items are given, 
;       stops tracing everything.
;    untrace-all
;       stops tracing everything
;    print-traceables list
;       prints the descriptions for the traceable items in list
;       if list is empty, prints for all traceable items

; Useful global variables and constants:
;    *traced-items*
;       the list of items currently being traced
;    *traceable-items*
;       the list of items which may be traced

(defvar *traced-items* nil "List of items currently being traced")
(defvar *traceable-items* nil "List of items which can be traced")

; traced-p returns t if item is being traced, nil otherwise
(defun traced-p (item) (member item *traced-items*))

; traceable adds items to the list of traceable-items - these items may then be
; traced with trace-item, and tested with traced-p

(defun traceable (item &optional description trace-values)
  (when description (setf (documentation item 'setf) description))
  (when trace-values (setp item 'TRACE-VALUES trace-values))
  (pushnew item *traceable-items*)
  )

; trace-item starts tracing on the items given - if no items are given, tracing 
; is turned on for all traceable items

(defun trace-item (&rest items)
  (if items
      (dolist (i items)
	      (if (member i *traceable-items*)
		  ; handle compound trace values
		  (if (getp i 'TRACE-VALUES)
		      (mapcar 'trace-item (getp i 'TRACE-VALUES))
		      (pushnew i *traced-items*))
		  (print-error i " is not a traceable item")
		  ))
      (trace-all))
  *traced-items*)

; trace-all turns on tracing for all traceable items

(defun trace-all ()
  (setq *traced-items* (remove-if #'(lambda (i) (getp i 'TRACE-VALUES))
				  *traceable-items*))
  )

; untrace-item stops tracing of things - if items are specified, tracing is
; turned off for them, otherwise all tracing is stopped

(defun untrace-item (&rest items) 
  (if items
      (dolist (i items)
	      ; handle compound trace values
	      (if (getp i 'TRACE-VALUES)
		  (mapcar 'untrace-item (getp i 'TRACE-VALUES))
	          (setq *traced-items* (remove i *traced-items*))))
      (untrace-all))
  *traced-items*)

; untrace-all turns off tracing for all traceable items
(defun untrace-all () (setq *traced-items* nil))

; print-traceables prints all traceables items or a selection of them
; given by the user. The item's name and description are printed.
 
(defun print-traceables (&rest items)
  (if items 
      (setq items (intersection items *traceable-items*))
      (setq items *traceable-items*))
  (dolist (item (if (listp items) items (list items)))
	  (print-line " "  (symbol-name item) ": " (documentation item 'setf))
	  ; if compound trace-item, print them
	  (if (getp item 'TRACE-VALUES)
	     (print-line "     Compound trace for: " (getp item 'TRACE-VALUES)))
	  )
  (print-line *nl* " Items currently being traced: " *traced-items*)
  )

; &rest args are the arguments to the function being called
#+lucid (defadvice (trace handle-quoted-traces) (&rest args)
  (let (tlist ilist)
       (dolist (a (cdar args))
	       (if (and (listp a) (equal (car a) 'quote))
		   (push (cadr a) ilist)
		   (push a tlist))
	       )
       (when ilist (apply 'trace-item ilist))
       (when (or tlist (null ilist))
       (apply-advice-continue (list (append (list 'TRACE) tlist) nil)))
  ))

;; allegro has "arglist" available - contains argumetns of call for 
; function, ((macro args) (..)) for macro
#+allegro (excl:advise trace :before handle-quoted-traces nil
		  (let* ((margs (car excl:arglist))
			 (mac (car margs))
			 (args (cdr margs)))
			(setq args (remove-if 'null 
			      (mapcar #'(lambda (a) (if (and (listp a) 
						    (equal (car a) 'quote)) 
					       (progn (trace-item (cadr a)) nil)
					       a)) args)))
			(setq excl:arglist (append (list (append (list mac) args) )
					      (cdr excl:arglist)))
			))


;; cmu doesn't have an advice facility, so I redefine trace and
;; untrace.  NOTE: the lucid and allegro versions do this by advising
;; the macro in such a way that trace-item gets called at macro
;; expansion time.  For consistency, I have done it that way here too,
;; even though it would seem more correct to do it at evaluation time.

#+cmu (defvar *redefined-trace-p* nil)
#+cmu (when (not *redefined-trace-p*)
	(setq *redefined-trace-p* t)
	(let ((old-trace (macro-function 'common-lisp:trace)))
	  (setf (macro-function 'common-lisp:trace)
	    (lambda (form env)
	      (let (quoted unquoted)
		(dolist (item (cdr form))
		  (if (and (listp item) (equal (car item) 'quote))
		      (push (cadr item) quoted)
		    (push item unquoted)))
		(if quoted
		    (apply #'trace-item quoted))
		(if unquoted
		    (funcall old-trace (cons 'trace unquoted) env)
		  nil))))))

; &rest args are the arguments to the function being called
#+lucid (defadvice (untrace handle-quoted-untraces) (&rest args)
  (let (tlist ilist)
       (dolist (a (cdar args))
	       (if (and (listp a) (equal (car a) 'quote))
		   (push (cadr a) ilist)
		   (push a tlist))
	       )
       (when ilist (apply 'untrace-item ilist))
       (when (or tlist (null ilist))
       (apply-advice-continue (list (append (list 'UNTRACE) tlist) nil)))
  ))

;; allegro has "arglist" available - contains argumetns of call for 
; function, ((macro args) (..)) for macro
#+allegro (excl:advise untrace :before handle-quoted-untraces nil
		  (let* ((margs (car excl:arglist))
			 (mac (car margs))
			 (args (cdr margs)))
			(setq args (remove-if 'null 
			      (mapcar #'(lambda (a) (if (and (listp a) 
						    (equal (car a) 'quote)) 
					       (progn (untrace-item (cadr a)) nil)
					       a)) args)))
			(setq excl:arglist (append (list (append (list mac) args) )
					      (cdr excl:arglist)))
			))

;; see note above for trace
#+cmu (defvar *redefined-untrace-p* nil)
#+cmu (when (not *redefined-untrace-p*)
	(setq *redefined-untrace-p* t)
	(let ((old-untrace (macro-function 'common-lisp:untrace)))
	  (setf (macro-function 'common-lisp:untrace)
	    (lambda (form env)
	      (let (quoted unquoted)
		(dolist (item (cdr form))
		  (if (and (listp item) (equal (car item) 'quote))
		      (push (cadr item) quoted)
		    (push item unquoted)))
		(if quoted
		    (apply #'untrace-item quoted))
		(if unquoted
		    (funcall old-untrace (cons 'untrace unquoted) env)
		  nil))))))
