;       Last change    August 1993
(in-package response)

; The build routines are used to create a noun phrase description for an
; entity by looking up the appropriate information in the knowledge
; base (like the type under tp.type for a physical object), or by formatting
; the information into a nice English form for quoted expressions (esp
; times and sets)

; if no infomation is found, or it is not usable, if the item is a variable
; use a noun based on its sort; if a constant use its name alone

(defun default-info (frag type &aux pp)
  (if (or (equal (fragment-objtype frag) 'variable)
	  (skolem-constant-p (fragment-name frag)))
      (if (equal (fragment-fragtype frag) 'NP)
	  (unless (or (fragment-determiner frag)
		      (and (setq pp (fragment-PP frag)) (null (fragment-prep pp))))
		  (add-noun (default-noun type) frag))
	  (setp frag 'NP (make-frag 'NP (cons 'name (fragment-name frag))
				    (cons 'head frag)
				    (cons 'noun (default-noun type)))) )
      (setp frag 'proper-noun (princ-to-string (fragment-name frag)))
      )
  frag)

; build episode tries to build an description for an episodic or propos constant
; It looks up the clauses under tp.description or tp.content, organizes 
; them, and then uses only the first of such clauses to try to build a sentence

(defun build-episode (argfrag sort)
  ; if still no info, use default info
  (if t
      (if (and (not (fragment-VP argfrag)) (not (fragment-NP argfrag))
	       (not (noun-info argfrag))
	       )
	      (default-info argfrag sort)
	  )
      )
  argfrag)

; the following routines attempt to print date information nicely. Dates are of
; the form (quote (time year month day hour minute second)), where any of the
; date fields may be unspecified (a concept instead of a number).

; date-part returns the three date fields of an absolute time expression

(defun date-part (timexp) 
  (subseq timexp '0 '3))

; date-part returns the three time fields of an absolute time expression

(defun time-part (timexp )
  (subseq timexp '3))

; find-month reeturns the character string for the month given 
; (so we can say in May instead of in month 05)

(defun find-month (mnum)
  (case (round mnum)
	(1 "January")
	(2 "February")
	(3 "March")
	(4 "April")
	(5 "May")
	(6 "June")
	(7 "July")
	(8 "August")
	(9 "September")
	(10 "October")
	(11 "November")
	(12 "December")
	(otherwise "some month")
	)
  )

; prepare-year sets up a PP fragment under frag for the year. If the year is
; a number, the PP will be "in year", otherwise, " in some unspecified year"

(defun prepare-year (year)
  (if (numberp year)
      (squish (list "in" " " (princ-to-string year)))
      "in some unspecified year")
  )

; if the month is not specified, but the day is, we need to say on the nth day
; of some unspecified month.
; nth-day returns a string corresponding to nth (only first,second, and third
; ar fully spelled out).

(defun nth-day (day)
  (case day
	(1 "first")
	(2 "second")
	(3 "third")
	(otherwise (squish (list (princ-to-string day) "th")))
	)
  )

; prepare-date prepares date information and subfragments under frag
; if all date fields are numbers, the date will be Month day, year (where
; the string for month replaces the number).
; Otherwise, at least one field is unspecified, so it gets a bit messier

(defun prepare-date (date &aux
			  (year (car date))
			  (month (cadr date))
			  (day (caddr date))
			  monthname)

  ; now handle month and day
  (if (numberp month)
      (setq monthname (find-month month)))
  (cond ((and monthname (numberp day))
	 ; both month and day specified
	 (if (numberp year)
	     ; if year specified, date is Month day, year
	     (squish (list monthname " " (princ-to-string day) ", "
			     (princ-to-string year)))
	     ; otherwide date is Month day in some unspecified year
	     (let ((y (prepare-year year)))
		  (squish (list monthname " " (princ-to-string day) " " y)))
	     ))
	; month specified, but not day
	((numberp month) ;some unspecified day in monthname
	 (let ((mpart 
		(if (numberp year) 
		    (squish (list monthname ", " (princ-to-string year)))
		    (squish (list monthname " in some unspecified year")))))
	      (squish (list "some unspecified day in " mpart))
	      ))
	; day specified, but not month
	((numberp day) ; the xth day in some unspecified month in 
		       ;  (some unspecifed) year
         (let* ((ypart (prepare-year year))
		(mpart " day in some unspecified month "))
	       (squish (list "the " (nth-day day) mpart ypart))
	      ))
	(t ; some unspecified day in some unspecified month in 
	   ;   (some unspecified) year
	   (let ((ypart (prepare-year year)))
		(squish (list "some unspecified day in some unspecified month "
			      ypart))) )
	 )
  )


; prepare time prepares the fragment frag with time information for the given 
; time
; if all fields are specified, the time is hh:mm:ss, if seconds are not hh:mm
; otherwise it is a bit messy

(defun prepare-time (time &aux
			  (hour (car time))
			  (minute (cadr time))
			  (second (caddr time))
			  hs ms ss)
  (when (numberp hour)
        (setq hs (princ-to-string hour)))
  (when (numberp minute)
	(if (< minute 10)
	    (setq ms (squish (list "0" (princ-to-string minute))))
	    (setq ms (princ-to-string minute))))
  (when (and (numberp second) (> second 0))
	(if (< second 10)
	    (setq ss (squish (list "0" (princ-to-string second))))
	    (setq ss (princ-to-string second))))
  (cond ((and (numberp hour) (numberp minute))
	 ; return hour:minute:second
	 (squish (if ss (list hs ":" ms ":" ss) (list hs ":" ms))))
	((numberp hour) ; minute unspecified - between that hour and the next
	 ; return NPor fragment hour:00 or (hour+1):00
	 (let* ((h2 (princ-to-string (+ hour 1))))
	       (squish (list hs ":00 or " h2 ":00"))))
	(t ; hour unspecified - some time
	   "some time"))
  )

; build-time takes a quoted expression for time, breaks it into its date and
; time constituents, and generates an exglish time expression from it.
; If the time is all 0's, only the date is returned
; Otherwise, the result in time on date

(defun build-time (argfrag &aux 
			   (fname (fragment-name argfrag))
			   (item (if (equal (term-type fname) 'record)
				     (record-contents fname)
				     (getp fname 'arguments))))
  (if (null item) 
      (default-info argfrag 'time)
      (let (date time) ; quoted expression

	   ; break up quoted expression into date and time
	   (setq date (date-part item))
	   (setq time (time-part item))

	   ; if time is all 0's, only put date stuff on argfrag
	   ; else put time stuff on argfrag, with PP on date
	   (if (equal time '(0 0 0))
	       ; only return date
	       (setp argfrag 'print (prepare-date date))

	       ; return time on date
	       (let* ((d (prepare-date date))
		      (tm (prepare-time time)))
		     (setp argfrag 'print (squish (list tm " on " d)))
		    ))
	   ))
  argfrag)

; build-number uses the string for the number if it is a number, and the
; default information otherwise

(defun build-number (argfrag)
  (if (null (getp (fragment-name argfrag) 'print))
      (default-info argfrag 'number)
      (if (numberp (fragment-name argfrag))
	  (setp argfrag 'print (princ-to-string (fragment-name argfrag)))
	  (let () ;quoted expression
	   (setp argfrag 'print 
	     (princ-to-string (cadr (getp (fragment-name argfrag) 
					  'print))))
	   )))
  argfrag)

; use-name decides on the name to print in a function description 
; - if a concept, its name, otherwise the argument itself

(defun use-name (arg)
      (princ-to-string arg)
  )

; build-function returns a description for the function given
; A description consisting if the function name
; followed by the argument names is used.

(defun build-function (argfrag &aux 
			       (fn (fragment-name argfrag))
			       (op (getp fn 'operator)))
  (cond ((equal op 'date) (build-time argfrag))
	((equal op 'set-of) (build-set argfrag))
	(t (when (null (noun-info argfrag))
		 (setp argfrag 'AP
		       (make-frag 'AP (cons 'name fn) (cons 'head argfrag)
				  (cons 'adjective 'function)))
		 (setp argfrag 'print
		       (squish (append (list (princ-to-string 
					      (getp fn 'operator)))
			     (apply 'append
			       (mapcar #'(lambda (a) (list " " (use-name a)))
					      (getp fn 'arguments))))))
		 )))
  argfrag)

; set-fragment returns an NP fragment with name name and head frag to
; use in building sets. Subject information is copied from the set head
; to the lower fragments

(defun set-fragment (name frag &aux
			  (nfrag (make-frag 'NP (cons 'name name)
				    (cons 'head frag)
				    (cons 'required t))))
  (set-subject (fragment-subject frag) nfrag)
  nfrag)

; build-set returns aN NPand fragment with an NP fragment beneath it for each
; member of the quoted set description.
; If not a quoted set, it returns the default information.

(defun build-set (argfrag &aux
			   (fname (fragment-name argfrag))
			   (item (if (equal (term-type fname) 'record)
				     (record-contents fname)
				     (getp fname 'arguments))))
  (if (null item) 
      (default-info argfrag 'set)

      ; otherwise it is a quoted expression, with everything there already
      (let ()
           ; put NP for each member under argfrag
           (setp argfrag 'parts 
		 (mapcar #'(lambda (a) (set-fragment a argfrag)) item))
           (setp argfrag 'conj 'and) 
           (setp argfrag 'number 'plural)
           )
      )
  ; check to see if new ones need to be filled here or are done automatically
  argfrag)

; build-NP fills in noun phrase information for quoted expressions and
; other entites. It first decides what type of entity it has (using sorts)
; and uses those to determine which filling routine to call.

(defun build-NP (argfrag &aux
			 (name (fragment-name argfrag))
			 (sort (if (equal name 'removed)
				   nil
				   (find-sort name))))
  ; first check to see if this is required

  (unless (or (null name)
	      (equal name 'removed)
	      (if (not (fragment-subject argfrag))
		  (noun-info argfrag) )
	      (fragment-proper-noun argfrag)
	      (fragment-print argfrag))  ; ie finished

	  ; use sort to determine what to do
	  (setq argfrag 
	  (cond ((getp name 'trans) 
		 (setp argfrag 'proper-noun (getp name 'trans))
		 argfrag)
		((equal sort 'EPISODE) (build-episode argfrag sort))
		((equal sort 'PROPOS) (build-episode argfrag sort))
		((equal sort 'TIME) (build-time argfrag))
		((and (equal sort 'SET) (not (fragment-noun argfrag)))
	         (build-set argfrag))
		((member sort '(INTEGER REAL NUMBER))
	         (build-number argfrag))
		(t (if (fragment-required argfrag)
		       (default-info argfrag sort)
		       argfrag))
		))
	  )
  argfrag)

