;       Last change    August 1993

(in-package response)

(tweakable '*default-lex* t 
  "Indicates whether to guess the lexical item (t), or ask the user (nil)"
  'boolean)

; Lexicon handlers

; actual verbs and nouns are treated by putting translations as a property
; on the symbol
; predicates have type and other information stored on the concept
; all verbs are their infinitives - other forms are either stored with the
; vrb infinitive name or calculated from it

; noun-properties returns the list of properties associated with a noun

(defmacro noun-properties (noun)
  `(get ,noun 'PROPS))

;determine noun looks up the appropriate form of the noun in question
; if not found, it prompts the user for it

(defun determine-noun (noun number &aux n nn)
  (if (not (symbolp noun)) (return-from determine-noun noun))
  (setq n
	(if (and (equal number 'plural)
		 (not (check-noun-property noun 'group)))
	    (getp noun 'PLURAL)
	    (string-downcase (princ-to-string noun))
	    ))

  ; ask user
  (when (null n)
	    (setq n (string-downcase (princ-to-string 
				      (if (equal number 'PLURAL)
					  (if (member 'MASS
						      (getp noun 'PROPS))
					      noun
					      (calc-word-form noun 's))
					  noun))))
	(if *default-lex*
	    n
	    (let ()
		(print-line "Missing lexical information for the noun form for "
				noun " " number)
		 (print-line "The default is " n)
		 (print-line "Please enter a string for the noun or () to accept the default: ")
		 (setq nn (read)) 
		 (when nn
		       (if (not (stringp nn))
			   (setq nn (string-downcase (prin1-to-string nn)))))
		 (if nn (setq n nn))) 
	    ))
  n)
	
(defun vowel-symbol (a)
  (member a '(A E I O U)))

(defun word-form (word ending)
  (string-downcase (princ-to-string (calc-word-form word ending))))

(defun calc-word-form (word ending &aux
				   (vstem (strip-pkg word)))
  (cond ((member ending '(ed ing))
	 (cond ((equal (last1 vstem) 'e)
		(setq vstem (remove 'e vstem 
				    :from-end t :count 1))
		(when (and (equal (last1 vstem) 'e) (equal ending 'ing))
		      (setq vstem (append vstem (list 'e))))
		(when (and (equal (last1 vstem) 'i) (equal ending 'ing))
		      (setq vstem (append 
				   (remove 'i vstem :from-end t :count 1)
				   (list 'y)))) )
	       ((and (equal (last1 vstem) 'y)
		     (not (vowel-symbol (nth (- (list-length vstem) 2) vstem)))
		     (equal ending 'ed))
		(setq vstem (append (remove 'y vstem 
				     :from-end t :count 1) (list 'i))))
	       ))
	((member ending '(s))
	 (cond ((and (equal (last1 vstem) 'y)
		     (not (vowel-symbol (nth (- (list-length vstem) 2) vstem))))
		(setq ending 'es)
		(setq vstem (append (remove 'y vstem 
					    :from-end t :count 1) (list 'i))))
	       ((member (last1 vstem) '(h s x z))
		(setq ending 'es))
	       ))
	)
  (build-symbol (append vstem (list ending)) *package*)
  )

(defun ends-in-s (str &aux (lst (- (car (array-dimensions str)) 1)))
 (member (geta str lst) '(#\s #\x #\z) :test 'equal)
 )

(defun pr-verb-form (verb tense number)
  (cond ((equal tense 'past) (word-form verb 'ed))
	((equal number 'plural) (string-downcase (princ-to-string verb)))
	(t (word-form verb 's))
	))

;determine verb looks up the appropriate form of the verb in question
; if not found, it prompts the user for it

(defun determine-verb (tense verb number negp given-aspect mood &optional is-aux &aux v nv
										      (plural (equal number 'plural))
										      (aspect
										       (cond ((listp given-aspect)
											      (if (member 'progressive given-aspect)
												  (if (member 'perfect given-aspect)
												      'perf-prog
												    'perfect)
												(car given-aspect)))
											     (t given-aspect)))
										      
										      )
  (if (not (symbolp verb)) (return-from determine-verb verb))
  (when (and mood (not (equal mood 'indicative)))
    (warn-error 
     "Response generator: Only indicative mood is currently supported" 
     *nl* "             - " mood " will be treated as indicative for now"))
  (if (stringp verb) (return-from determine-verb verb))
  
  ;; look it up
  (setq v
    (case aspect
      (infinitive 
       (squish (list "to " (string-downcase (princ-to-string verb)))))
      (prespart (get verb 'PRESPART))
      (pastpart (get verb 'PASTPART))
      (passive (get verb 'PASSIVE))
      (progressive 
       (squish (list (determine-verb tense 'be number negp nil nil t)
		     " " (determine-verb nil verb number nil 'prespart nil)))
       )
      (perfect 
       (squish (list (determine-verb tense 'have number negp nil nil t)
		     " " (determine-verb nil verb number nil 'pastpart nil)))
       )
      (perf-prog
       (squish (list (determine-verb tense 'have number negp nil nil t)
		     " " (determine-verb nil 'be number nil 'pastpart nil t)
		     " " (determine-verb nil verb number nil 'prespart nil)))
       )
      (t 
       (if negp
	   (if plural 
	       (case tense
		 (past (cadr (getp verb 'NEGATIVE-PAST)))
		 (future (squish (list 
				  (determine-verb 'present 'will number negp 
						  nil nil) " " 
						  (string-downcase (princ-to-string verb)))))
		 (otherwise (cadr (getp verb 'NEGATIVE-PRESENT))))
	     (case tense
	       (past (car (getp verb 'NEGATIVE-PAST)))
	       (future (squish (list 
				(determine-verb 'present 'will number negp 
						nil nil) " " 
						(string-downcase (princ-to-string verb)))))
	       (otherwise (car (getp verb 'NEGATIVE-PRESENT)))))
	 (if plural
	     (case tense
	       (past (cadr (getp verb 'PAST)))
	       (future (squish (list 
				(determine-verb 'present 'will number negp 
						nil nil) " " 
						(string-downcase (princ-to-string verb)))))
	       (otherwise (cadr (getp verb 'PRESENT))))
	   (case tense
	     (past (car (getp verb 'PAST)))
	     (future (squish (list 
			      (determine-verb 'present 'will number negp 
					      nil nil) " " 
					      (string-downcase (princ-to-string verb)))))
	     (otherwise
	      (car (getp verb 'PRESENT)))))
	 ))
      ))
  ;; ask user
  (when (null v)
    (setq v (calc-default-verb tense verb number negp aspect is-aux)) 
    (if *default-lex*
	v
      (let ()
	(print-line "Missing lexical information for the verb form for " 
		    verb " " (if tense tense 'present)
		    " " (if number number 'singular) " "
		    (if negp "negated" " ") " " 
		    (case aspect
		      (infinitive "infinitive")
		      (passive "passive")
		      (prespart "present participle (ing)")
		      (pastpart "past participle (ed)")
		      (otherwise " ")))
	(print-line "The default form is " v )
	(print-line "Please enter a string for this verb or () to accept the default: ")
	(setq nv (read))
	(when nv 
	  (if (not (stringp nv))
	      (setq nv (string-downcase (prin1-to-string nv)))))
	(if nv (setq v nv))
	))
    )
  v)

(defun calc-default-verb (tense verb number negp aspect is-aux) 
  (case aspect
    (prespart (word-form verb 'ing))
    (pastpart (word-form verb 'ed))
    (passive (word-form verb 'ed))
    (otherwise 
     (if negp
	 (if is-aux 
					; get proper form of verb and append not
	     (squish (list 
		      (determine-verb tense verb number nil nil nil) " not"))
					; choose proper form of do and append verb to it
	   (squish (list
		    (determine-verb tense 'do number negp nil nil) 
		    " " (string-downcase (princ-to-string verb))))
	   )
       (pr-verb-form verb tense number)))
    )
  )

; adj-trans looks up the string form for a given adjective, prepostion, etc
; if not found, it converts the name to a lower case string and uses that

(defun adj-trans (adj &aux a)
  (setq a (get adj 'TRANS))
  (when (null a)
	(setq a (substitute (char " " 0) (char "-" 0)
			    (string-downcase (princ-to-string adj))))
        )
  a
  )

; get-pronoun returns any pronouns associated with a name

(defun get-pronoun (name &optional subject)
  (when (symbolp name)
  (if subject 
      (getp name 'SUBJECT-PRONOUN)
      (getp name 'OBJECT-PRONOUN)))
  )

; prep-adv-trans calls adj-trans to get a string for an adverb or preposition

(defun prep-adv-trans (item)
  (adj-trans item))

(set-display-function 'lex 'lexical-entry "Shows the lexical information for a particular object")

; pr-commas prints a list of verb lexical entries separated by commas

(defun pr-commas (llist &aux
			nlist)
  (dolist (l llist)
	  (if (null nlist)
	      (setq nlist (list l))
	      (setq nlist (append nlist (list "," l))))
	  )
  nlist)

; lexical-entry prints the lexical information for the arguments given

(defun lexical-entry (brief &rest args)
  (declare (ignore brief))
  (dolist (a args)
	  (dolist (type (getp a 'LEXICAL-TYPE))
		  (case type 
			(ADJ
			 (print-line a #\tab " Adjective: "
				    (getp a 'TRANS)))
			(NOUN
			 (print-line a #\tab " Noun: Singular: " a
				    #\tab " Plural: " (getp a 'PLURAL) )
			 (when (getp a 'PROPS)
			       (print-line #\tab " Properties: " 
				    (getp a 'PROPS))))
			(NAME
			 (print-line  a #\tab " Name: " (getp a 'TRANS)
				    #\tab " Pronouns: " 
				    (list (getp a 'SUBJECT-PRONOUN) 
					  (getp a 'OBJECT-PRONOUN))))
			(VERB
			 (print-line a #\tab " Verb: Infinitive: " a 
			    #\tab " Passive: " (get a 'PASSIVE) 
			    #\tab " Present Participle: " 
				(getp a 'PRESPART)
			    *nl* #\tab "    Positive Forms: " 
			    "present " (get a 'PRESENT)
			    " past " (get a 'PAST)
			    *nl* #\tab "    Negative Forms: " 
			    "present " (get a 'NEGATIVE-PRESENT)
			    " past " (get a 'NEGATIVE-PAST)
				       ))
			(OTHER
			 (print-line a #\tab " Adverb/Preposition/Other: " 
				    (getp a 'TRANS)))
			)))
  )

(defun add-lex (word type &key
		     ; verbs
		     present past negative-present negative-past passive
		     prespart pastpart
		     ;nouns
		     plural props
		     ;names
		     trans subject-pronoun object-pronoun
		     )
  (setp word 'LEXICAL-TYPE type)
  (case type 
	(VERB 
	 (setp word 'PRESENT present)
	 (setp word 'PAST past)
	 (setp word 'NEGATIVE-PRESENT negative-present)
	 (setp word 'NEGATIVE-PAST negative-past)
	 (setp word 'PASSIVE passive)
	 (setp word 'PRESPART prespart)
	 (setp word 'PASTPART pastpart)
	 (if (and pastpart (not passive)) 
	     (setp word 'PASSIVE pastpart)
	     (if (and passive (not pastpart)
		      (not (equal passive 'no-passive)) )
		 (setp word 'PASTPART passive)))
	 )
	(NOUN
	 (setp word 'PLURAL plural)
	 (setp word 'PROPS props)
	 )
	(NAME
	 (setp word 'TRANS trans)
	 (setp word 'SUBJECT-PRONOUN subject-pronoun)
	 (setp word 'OBJECT-PRONOUN object-pronoun)
	 (setp word 'PROPS props)
	 )
	(OTHER
	 (setp word 'TRANS trans)
	 )
	)
  )
