;       Last change    August 1993
(in-package spec)
;                  Specialist Interface
;
; These routines include routines for setting up and activating specialsits,
; as well as for invoking them during the operation of the theorem prover.
;
;  Activation:
;    use-spec specs
;     activates all specialist in the specs list
;  Setting up:
;    set-specialist name directory nicknames
;     tells the system that specialist name is available, and 
;     located in directory
;    define-specialist spec loadfile preds sorts functions operators topic
;     used in the activation of a specialist to indicate to the system
;     what predicates and fucntions are involved, and what sorts are allowd
;     for the first argument.
;
; Useful global varialbes and constants:
;   *maximum-effort*  - the highest effort value that may be used in the system
;   *available-specialists* - a list of the specialist which are available
;   *active-specailists*  - a list of the specialists which are currently
;      active
;   *current-netname* - the current subnet path

; Controls on the interface:
;   *specialist-entry-effort* - effort to use by specialists in consistency 
;      checking
;   *specialist-eval-effort* - effort to use by specialists in evaluation 
;      of literals and comparison of literals
;   *spec-enter* - flag indicating whether literals should be entered into
;      specialists domains
;   *spec-eval* - flag indicating whether specialists should be used to
;      evaluate wffs
;   *spec-compare-preds* - flag indicating whether specialists should be used
;      to compare predicates
;   *spec-compare-lits* - flag indicating whether specialists should be used
;      to compare literals
;   *spec-assumptions* - flag indicating whether assumptions should be 
;      temporarily loaded into specialists domains

(traceable 'function-eval "Traces function evaluations")
(traceable 'interested-party "Traces addition to interested party lists")
(traceable 'spec-eval "Traces use of specialsist for evaluation")
(traceable 'spec-entry "Traces entry into specialist domains")
(traceable 'spec-test "Traces comparisons")
(traceable 'spec-assert "Traces assertions made by specialists")
(traceable 'spec-int "Traces specialist interface operations"
	   '(function-eval interested-party spec-eval spec-entry spec-test
			   spec-assert))

(defvar *current-netname* '/ "Current subnet path")
(defvar *current-netnames* '/ "Current subnet paths")
(defvar *current-embedding* 'embedded "Current subnet path")
(defvar *current-literal* nil "Current literal being entered")
(defvar *flattened-lit* nil "Flattened version of literal")

; spec-compare-preds uses the specialists to help determine the relationship between
; two predicates. Tis is then used to determine whehter the predicates are
; compatible or incompatible.
; Note - this may change to have the specialists directly responsible for
; determining compatibility/incompatibility of predicates if probabilities
; get involved.

(defun spec-compare-preds (gpred1 gpred2 &aux 
			     spec1 spec2 ans fn)
  (cond ((equal gpred1 gpred2) 'equivalent)
	((or (lambda-pred-p gpred1) (lambda-pred-p gpred2))
	 'unknown)
	((and (numberp gpred1) (numberp gpred2))
	 (if (<= gpred1 gpred2) 'equivalent 'unknown))
	((and (symbolp gpred1) (symbolp gpred2)
	      (setq spec1 (specs-for-pred-or-op gpred1))
	      (setq spec2 (specs-for-pred-or-op gpred2)))
	 (dolist (s (intersection spec1 spec2))
		 (when (setq fn (find-symbol "COMPARE-PREDS" s))
		       (setq ans (apply fn (list gpred1 gpred2)))
		       (unless (equal ans 'UNKNOWN)
			       (return ans))
		       ))
	 (if ans ans 'UNKNOWN)
	 )
	(t 'UNKNOWN))
  )


(defvar *active-specialists* nil
      "Names of currently active specialists")

(defvar *available-specialists* nil
      "Names of available specialists" )

; set-specialist tells the net that a particular specialist is available,
; located in a particular directory, and may be activated by its name or
; any of a number of synonyms

(defun set-specialist (sym defn-file &optional nicknames description) 
    (setp sym 'definition defn-file)
    (when description (setf (documentation sym 'setf) description))
    (dolist (n nicknames) (setp n 'spec-name sym))
    (pushnew sym *available-specialists*)
    )

(set-display-function 'active-specs 'active-specialists 
  "Displays currently active specialists")
(set-display-function 'avail-specs 'available-specialists 
  "Displays specialists which may be activated")
(set-display-function 'specs 'available-specialists 
  "Displays specialists which may be activated")

; active-specialists displays the list of currently active specialists
 
(defun active-specialists (&optional brief &rest args)
  (declare (ignore args))
  (print-line "Specialists currently active are: ")
  (if brief
      (mapcar #'(lambda (s) (print-line "     " s)) *active-specialists*)
      (mapcar #'(lambda (s) 
			(print-line "     " s)
			(print-line "           " (documentation s 'setf))) 
	      *active-specialists*) )
  nil)
     
; available-specialists displays the list of available specialists
(defun available-specialists (&optional brief &rest args)
  (declare (ignore args))
  (print-line "Specialists available are: ")
  (if brief
      (mapcar #'(lambda (s) (print-line "     " s)) *available-specialists*)
      (mapcar #'(lambda (s) 
			(print-line "     " s)
			(print-line "           " (documentation s 'setf)))
	      *available-specialists*))
  nil)

; set-arg-sorts saves the list of allowed sorts for the 1st argument of
; literals that the given specialist looks at

(defun set-arg-sorts (specname sorts)
  (setp specname 'first-sorts sorts)
  )

; define-specialist defines a specialist to the system (predicates, functions,
;; and sorts), and loads the specialist - done only during activation
 
(defun define-specialist (specname specload preds 
				   &optional sorts functions operators topic)
  (unless (null specload)
	  (require specname specload)
	  )
  (when topic 
	(setp specname 'indicators (list topic))
	(setp topic 'specialists (append (getp topic 'specialists)
					 (list specname)))
	)
  (mapcar #'(lambda (p) (create-predicate p specname)) preds)
  (mapcar #'(lambda (p) (create-function p specname)) functions)
  (mapcar #'(lambda (p) (create-operator p specname)) operators)
  (set-arg-sorts specname sorts)
  (pushnew specname *active-specialists*)
  )

; load-spec activates a specialist - it first checks to see if the specialist
; is already active.
 
(defun load-spec (spec &aux lres)
  (if (member spec *active-specialists*)
      (print-line spec " is already active")

      (let ((filename (getp spec 'definition)))
		
	   ; check to see if there is a definition file there
	   ; if so, load it, else return message
	   (setq lres (load filename :if-does-not-exist nil))
	   (if (null lres)
	       (print-line "Specialist " spec " is not available")
	       (let ()
	       (print-line "Specialist " spec " is now active")
	       spec)
	       )
	   ) )
  )

; use-spec activates the given specialist arguments. If any have already been
; activated, or are not available, a message is printed.

(defun use-spec (&rest args &aux spec) 
  (rewind-question)
  (cond ((not args)
	 (print-line "Specialists remaining which can be activated:")
	 (mapcar #'(lambda (s) 
			(print-line "     " s)
			(print-line "           " (documentation s 'setf))) 
	      (set-difference *available-specialists* *active-specialists*) )
	 nil)
	(t (mapcar #'(lambda (arg)
			(if (member arg *available-specialists*)
			    (setq spec arg)
			    (setq spec (cond ((getp arg 'spec-name)) (t arg))))
			(if (member spec *available-specialists*)
			    (load-spec spec)
			    (print-line spec " is not an available specialist"))
			)
	      args) ) )
  )

(defstruct interested-party
  (netname nil)				; context to enter in
  (lit nil)				; literal to be reasserted
  (spec nil)				; specialist interested
  )

; prep-arg prepares an argument name. If there is a sort, it is appended
; with an _ to the name

(defun prep-arg (arg)
  (if (symbolp arg)
      (let* ((name arg)
             (sort (find-sort arg)))
            (if (null sort)
                name
                (pack (list name '_ sort))
                )
            )    
      arg)
  )

; combine-context prepares a lits of the form 
;      (john believes (mary believes !embed-prop))
; where !embed-prop is a special symbol used to indicate where the next 
; level proposition should go

(defun combine-context (context arg pred &aux
				(argitem (prep-arg arg))
				(newprop (list argitem pred '!embed-prop)))
  (if (null context)
      newprop
      (subst newprop '!embed-prop context :test 'equal)
      )
  )

(defun combine-names (nlist)
  (cond ((null nlist) nil)
	((atom nlist) (list '/ nlist))
	(t (append (list '/ (car nlist)) (combine-names (cdr nlist))))
	)
  )

; strip-prop determines what the literal and subnet context should be for
; the current assertion. If we have a wff of the form
;  ((john believe p1) * e1), we pick off the sunet information from the
; modal context (john), and look up the content of p1 to find out what the
; actual literal should be.
; Otherwise the subnet is the main one (/), and the literal is the
; assertion itself.
; Strip-prop returns a list consisting of the subnet name, the literal to
; be sent to the specialists, and the predicate involved (in case we ever
; want to differentiate based on that)

(defun modal-prop (netnames prop pred context embedding &aux
			   (args (getp prop 'arguments))
			   (wff (last1 args))
			   (savewff wff)
			   diff)
	 
  ; check #arguments - if more than 2, last one is the desired new wff
  (cond ((null wff) nil)
	((equal (getp wff 'entity-type) 'term)
	 (case (term-type wff)
	       (constant
		(cond ((getp wff 'print) (setq wff (getp wff 'print)))
		      ((equal (find-sort wff) 'propos)
		      ; find content, use that
		      (setq diff t)
		      (let ((ws (get-class (list wff '_ 'object))))
			   (if ws 
			       (setq wff (car (getp (select-one ws) 'arguments)))
			       (setq wff nil))
		      )) ))
	       (modified
		(if (equal (getp wff 'operator) 'that)
		    (setq wff (car (getp wff 'arguments)))
		    ; what else can be done here?
		    ))
	       (wff nil)
	       (variable (setq wff nil))
	       (otherwise (setq wff nil))))
	((equal (getp wff 'entity-type) 'wff) nil)
	(t (setq wff nil))
	)
  (cond (wff
	 (setq netnames (make-all-netnames netnames (remove savewff args) pred))
	 (setq context (if (listp (car context)) 
			   (append context (list (car args) pred 'subject))
			   (list context (list (car args) pred 'subject))))
	 (unless diff
	 (setq embedding (subst embedding (print-info wff) (print-info prop)
				:test 'equal)))
	 (strip-prop netnames wff context embedding) )
	(t (list netnames prop context embedding))
	)
  )

(defun make-all-netnames (netnames args pred &aux
				   nlist arglists)
  (dolist (n netnames)
	  (setq arglists (mapcar #'(lambda (a) (cond ((get-equivalence a n))
						     (t (list a)))) args))
	  (setq nlist (append nlist (make-netnames n nil arglists pred)))
	  )
  nlist)

(defun select-one (list-of-wff)
  (dolist (w list-of-wff)
	  (if (equal (getp w 'pred) '_)
	      (return-from select-one w))
	  )
  )

(defun strip-prop (netnames prop context &optional (embedding 'embedded) &aux
			   pred)
		 
  (case (wff-type prop)
	(prefix
	 (if (subnet-pred (setq pred (getp prop 'pred)))
	     ; actual wff is determined by the second or third argument
	     ; context determine by other arguments and predicate
	     (modal-prop netnames prop pred context embedding)
	     (list netnames prop context embedding))
	 )
	(episodic
	 (if (or (equal (getp prop 'pred) '_)
		 (and (equal (getp (car (getp prop 'arguments)) 'entity-type) 'term)
		      (equal (getp (car (getp prop 'arguments)) 'term-type) 'function)))
	     (list netnames prop context embedding)
	     (let ((wff (car (getp prop 'arguments))))
		  ;building in mp here - any inference from a ** should use *
		  (strip-prop netnames wff context
			 (subst '* '** 
			    (subst embedding (print-info wff) (print-info prop)
				:test 'equal))))
	     )
	 )
	(logical (list netnames prop context embedding))
	(causal (list netnames prop context embedding))
	(otherwise (list netnames prop context embedding))
	)
  )

; strip-modals rmoves all modal context embedding the time literal to
; be entered/evaluated. It returns a list consisting of the subnet
; identifier and the final proposition.

(defun strip-modals (literal)
  (strip-prop '(/) literal nil 'embedded)
  )

; unify-lists unifies two argument lists in the order given by ordlist.
; Testfn is a function passed to unify to determine if two constants should 
; unify

(defun unify-lists (arglist1 arglist2 &optional ordlist testfn &aux
			     (l1 nil)
			     (l2 nil))
  (unless ordlist
	  (setq l1 arglist1)
	  (setq l2 arglist2))
  (dolist (o ordlist)
	  (let ((pos1 (car o))
		(pos2 (cdr o)))
	       (setq l1 (append l1 (list (nth (- pos1 '1) arglist1))))
	       (setq l2 (append l2 (list (nth (- pos2 '1) arglist2)))))
	  )
  (compare-arglists l1 l2 :prefer 1 :match-const-fn testfn :cleanup t)
  )

; replace-prop replaces the arguments in a proposition by new (simplified) ones
(defun replace-prop (prop newargs &aux
                          (oldargs (getp prop 'arguments))
                          (sublist (arg-sub-list newargs oldargs)))
  (if (or (not *spec-simplify*) (equal oldargs newargs))
      prop
      (normalize (make-spec-inference (substitute-prop prop sublist)) '+ nil nil))
  )

; replace-function replaces the arguments to a function by new (simplified) ones
(defun replace-function (fn newargs &aux
                            (oldargs (getp fn 'arguments)))
  (if (equal newargs oldargs)
      fn 
      (let ((newf (new-symbol 'term)))
           (setp newf 'operator (getp fn 'operator))
           (setp newf 'arguments newargs)
           (setp newf 'entity-type 'term)
           (setp newf 'term-type 'function)
           newf))
  )

; spec-eval-fn evaluates an argument which is a function. 
; This is done recursively, and no variables allowed!

(defun spec-eval-fn (arg &aux
		    (fn-name (getp arg 'operator))
		    (specs (if (symbolp fn-name) (getp fn-name 'SPECIALISTS)))
		    (args (mapcar 'evalarg (getp arg 'arguments)))
		    p)

  (if (traced-p 'function-eval)
      (print-line "  Specialists: Evaluating function " fn-name " " args) )
  (unless (null args)
    (if (non-constant-arg args)
	(replace-function arg args)
      (let ((ans nil))
	(dolist (s specs)
	  (let ((fnsym (if (equal s 'other-specialist)
			   (if (setq p (getp fn-name 'package))
			       (find-symbol (symbol-name fn-name) p))
			 (find-symbol (symbol-name fn-name) s))) )
	    (unless (null fnsym)
	      (setq ans (apply fnsym args)))
	    (unless (or (null ans)
			(equal ans 'unknown))
	      (return ans))
	    )
	  )
	(if ans
	    ans
	  (replace-function arg args)))
      )
    )
  )

; evalarg evaluates an argument - if it is a function, the function is
; evaluated, otherwise the concept itself is returned

(defun evalarg (arg) 
  (cond ((symbolp arg) 
	 (case (getp arg 'term-type)
	       (function (spec-eval-fn arg))
	       (modified (spec-eval-fn arg))
	       (otherwise arg)))
	(t arg))
  )


; spec-eval-args gets and evaluates all arguments for a proposition
; a list of the evaluated arguments as concepts is returned

(defun spec-eval-args (p &aux
		  (arglist (getp p 'arguments))); (prop-args p)
  (mapcar 'evalarg arglist)
  )

; eval-using-spec checks the 1st argument of the literal agasint the sorts
; allowed for the specialist. If there is a match, that specialist evaluate
; routine is called.

(defun eval-using-spec (spec netnames lit negp pred arglist effort &aux
			     (speceval (find-symbol "EVALUATE" spec))
			     (specsorts (getp spec 'first-sorts))
			     res)
  (if (null speceval)
      (setq res 'unknown)
      (if (or (null specsorts)
	      (member (find-sort (car arglist)) specsorts :test 
		      #'(lambda (a b) (compatible-sorts b a t)))
	      )
	  (let ()
	       (setq *current-netname* (car netnames))
	       ;; note that (cdr netnames) is ignored.  This isn't my
	       ;; doing--it was like that already, contrary to what
	       ;; the docs say. -ANK
	       (setq res (apply speceval (list (car netnames) lit negp 
					       pred arglist effort))))
	  (setq res 'unknown)
	  )
      )
  res)

; this isn't probably the right thing to do
(defun get-spec-pred (pred)
  (cond ((equal (getp pred 'pred-type) 'modified)
	 (car (getp pred 'arguments)))
	(t pred))
  )

; eval-with-spec decides whether or not a specialist can be used to
; evaluate the given proposition. If so, it is called.

(defun eval-with-spec (netnames useprop &optional effort assumptions &aux
			       (prop useprop) ;(last1 (flatten useprop))
			       (pr (getp prop 'pred))
			       (negp (neg-p prop)) 
			       (specs (specs-for-pred-or-op pr))
			       (arglist (spec-eval-args prop))
			       (ans 'unknown)
			       cp)
  (if (null effort) (setq effort *specialist-eval-effort*))
  (if (symbolp netnames) (setq netnames (list netnames)))
  (unless (or (null arglist)
	      (null pr))

	  (when (and (traced-p 'spec-eval) specs)
	        (print-line "  Specialists: Evaluating " (getp prop 'print))
		(unless (equal (car netnames) '/)
			(print-it "    in subnet " (car netnames)))
		)
	  ; checkpoint and enter assumptions into spec domains
	  (when (and assumptions specs *spec-enter* *spec-assumptions*)
		(setq cp (checkpoint "Spec-Eval"))
		(dolist (a assumptions) (spec-enter a)) 
		)
	  (dolist (s specs)
		  (setq ans (eval-using-spec s netnames prop negp 
					     pr arglist effort)) 
		  (unless (equal ans 'unknown) 
			  (return ans))
		  )
	  ; retract assumptions
	  (when (and specs assumptions *spec-enter* *spec-assumptions*) 
		(retract cp :remove t))

	  ; if the answer is unknown, return simplified version if args were
	  ; evaluated

	  (when (equal ans 'unknown)
	        (setq ans (replace-prop prop arglist))
	        (if (equal ans prop)
		(setq ans 'unknown))
	        )
	  )
  ans)

; spec-evaluate tries to evaluate the given proposition using a specialist.
; Any modal embedding is stripped off and then eval-with-spec is used to
; determine the appropriate specialist(s).
;
; The effort field is optional and indicates how hard the specialist should
; work at evaluating the proposition. If not given, it defaults to the
; global specialist effort indicator.

(defun spec-evaluate (literal &optional effort assumptions &aux
		         (netprop (strip-modals literal))
		         (netnames (car netprop))
		         (prop (cadr netprop))
		         (pred (getp prop 'pred)) 
			 (result 'unknown)
			 (depth effort))
  (setq *current-netnames* netnames)
  (setq *current-literal* prop)
  (setq *current-embedding* (cadddr netprop))
  (setq *flattened-lit* nil)
  (if (null depth) (setq depth *specialist-eval-effort*))
  (unless (null prop)

	  ; if or, not, nec, cannot answer (and will never come here)

	  (if (or (member pred '(and or implies ~> <=>))
		  (numberp pred)
		  (member (getp prop 'operator) '(nec poss)))
	      (setq result 'unknown)

	      ; otherwise see if a specialist can answer

	      (let ()
		   (setq result (eval-with-spec netnames prop depth assumptions))
		  )
	 
	      )
	  )
  (convert-result result)
  )

; enter-using-spec checks the 1st argument of the literal against the
; sorts for the given specialist. If there is a match, the enter routine for
; that specialist is called.

(defun enter-using-spec (spec netnames lit negp pred arglist &aux
			      (oldcur *current-literal*)
			      res
			      (specenter (find-symbol "ENTER" spec))
			      (specsorts (getp spec 'first-sorts))
			      )
  (unless (null specenter)
	  (when (or (null specsorts)
		    (member (find-sort (car arglist)) specsorts :test 
		      #'(lambda (a b) (compatible-sorts b a t))))
		(setq *current-literal* lit)
		(dolist (n netnames)
			(setq *current-netname* n)
			(setq res (apply specenter 
				 (list n lit negp pred arglist))) )
		(setq *current-literal* oldcur)
		res)
	  )
  )

(defun reassert-to-spec (arg int-party &aux
			   (spec (interested-party-spec int-party))
			   (lit (interested-party-lit int-party))
			   (netname (interested-party-netname int-party))
			   (negp (neg-p lit))
			   (pr (getp lit 'pred))
			   )
  (when (and (member netname *current-netnames*)
	     (not (equal lit *current-literal*))
	     (not (equal lit *flattened-lit*))
	     )
	(if (traced-p 'interested-party)
	    (print-line *nl*
	     "  Specialist interface: Reasserting from interested party list of " 
	     arg *nl* *spaces*
	     "          " spec " " netname " " lit ": " (getp lit 'print)))
        (enter-using-spec spec (list netname) lit negp pr (spec-eval-args lit))) 
  ) 

(defun enter-interested (arg &aux 
			     (int-partys (if (symbolp arg)
			      (getp arg 'INTERESTED-PARTY)
			      nil)))
  (unless (null int-partys)
	  (mapcar #'(lambda (a) (reassert-to-spec arg a)) int-partys))
  )

(defun specs-for-item (item &aux slist news)
  (setq slist (getp item 'specialists))
  (dolist (tp (getp item 'indicators))
	  (if (setq news (getp (if (listp tp) (car tp) tp) 'specialists))
	      (setq slist (union slist news))
	      ))
  slist)

(defun specs-for-pred-or-op (item)
  (when (and item (symbolp item))
	(case (entity-type item)
	      (pred
	       (case (pred-type item)
		     (modified
		      (union (specs-for-pred-or-op (getp item 'operator))
			     (apply 'append (mapcar 'specs-for-pred-or-op
						    (getp item 'arguments)))))
		     (otherwise (specs-for-item item)) ))
	      (operator
	       (case (op-type item)
		     (modified
		      (union (specs-for-pred-or-op (getp item 'operator))
			     (apply 'append (mapcar 'specs-for-pred-or-op
						    (getp item 'arguments)))))
		     (otherwise (specs-for-item item))))
	      ))
  )

(defun interested-specs (wff &aux
			     (wff-type (getp wff 'wff-type))
			     sp specs)
  (case wff-type
	((prefix causal episodic)
	 (specs-for-pred-or-op (getp wff 'pred)))
	((quantified logical)
	 (dolist (k (append (getp wff '+trigger) (getp wff '-trigger)
			    (getp wff '+other) (getp wff '-other)))
		 (setq sp (interested-specs k))
		 (if sp
		     (if specs
			 (cond ((setq specs (intersection specs sp)))
			       (t (return)))
			 (setq specs sp))
		     (return)) )
	 specs)
	(otherwise nil)
	)
  )



; enter-with-spec decides whether or not a specialist would be interested
; in this proposition. If so, it is called to enter it.

(defun enter-with-spec (netnames prop &aux
				(pr (getp prop 'pred))
				(negp (neg-p prop)) 
				(specs (specs-for-pred-or-op pr))
				(arglist nil)
				(res nil))
  (setq *changed-concepts* nil)
  (unless (or (null specs)
	      (null pr))
	      (when (traced-p 'spec-entry)
		    (print-line "  Specialists: Entering " (getp prop 'print))
		    (unless (equal (car netnames) '/)
			    (print-it "    in subnet(s) " netnames)) )
	  (setq arglist (spec-eval-args prop))
	  (unless (null arglist)
		  (setq *current-literal* prop)
	          (setq res (or (mapcar #'(lambda (s)
			     (enter-using-spec s netnames prop   
					       negp pr arglist)) specs)))
		  (when res
		      (mapcar 'enter-interested arglist)
		      (do ((citem (pop *changed-concepts*) 
				  (pop *changed-concepts*)))
			  ((null citem))
			  (enter-interested citem)
			  ))
		  res)
	  )
  )


; spec-enter is called when a proposition is entered to have it entered into
; applicable specialists as well. If the top level predicate is AND, each
; proposition argument is entered, recursively. If or, not or nec, nothing is
; done (no specialists work with those). Otherwise, if there is any modal
; embedding, it is stripped off.
; The final proposition is then evaluated to see if it would be inconsistent
; to add it. If not, enter-with-spec is called to determine interested 
; specialists.

(defun spec-enter (literal &aux
		      (netprop (strip-modals literal))
		      (netnames (car netprop))
		      (prop (cadr netprop))
		      (pred (getp prop 'pred))) 
  (setq *current-netname* netnames)
  (setq *current-embedding* (cadddr netprop))
  (setq *flattened-lit* nil)
  (spec-enter-formula netnames prop pred)
  (check-subnet-equal)
  )
  
(defun spec-enter-formula (netnames prop pred)
  (unless (null prop)
	  ; if or, not, nec, do nothing (and won't get here)
	  (unless (or (member pred '(and or implies ~> <=>))
		  (numberp pred)
		  (member (getp prop 'operator) '(nec poss)))
	      ; replace things like ((cardinality-of c) < 3) with
	      ; 2 phrases (c has-cardinality d) and (d < 3)
	      (mapcar #'(lambda (p) (enter-with-spec netnames p))
		      (flatten prop))
	      ))
  )

; flatten wffs using functions into several using relational preds

(defun flatten (prop &aux
			 (pred (getp prop 'pred))
			 (args (getp prop 'arguments))
			 change newargs 
			 newprops
			 rp f)
  (when pred
	(dolist (a args)
		(if (and (symbolp a)
			 (equal (getp a 'term-type) 'function)
			 (symbolp (setq f (getp a 'operator)))
			 (setq rp (getp f 'REL-PRED)))
		    (let* ((fargs (getp a 'arguments))
			   (cname (pack (cons f fargs))) 
			  p1)
			 (setq change t)
			 ; prepare new constant, replace arg with that
			 ; cname needs a sort
			 (setp cname 'SORT (getp f 'SORT))
			 (when (epi-variable (car fargs))
			       (setp cname 'VARIABLE t)
			       (setp cname 'MATCHABLE 
				     (getp (car fargs) 'MATCHABLE))
			       )
			 (setq newargs (append newargs (list cname)))
			 ; prepare new prop using rel-pred
			 (setq p1 (normalize (append (list (car fargs) rp)
						     (cdr fargs)
						     (list cname)) '+ nil nil))
			  (setq *flattened-lit* p1)
			 (setq newprops (append newprops (list p1)))
			 )
		    (setq newargs (append newargs (list a))))
		)
	(when change
	    ;prepare list of props
	    (setq newprops (append 
		  (list (normalize (append (if (neg-p prop) (list 'not))
					   (list (car newargs) pred)
					   (cdr newargs)) '+ nil nil))
			    newprops))
	      ))
  (if newprops
      newprops
      (list prop))
  )


; convert-result converts a nil (false) result to no, and a true one to yes

(defun convert-result (result)
  (cond ((null result)
	 'no)
	((equal result t)
	 'yes)
	((equal result 'unknown)
	 'unknown)
	(t result)
	)
  )

; resolve-using-spec calls the resolve or factor (depending in flag resolve)
; routine of the specialist spec with the given literla

(defun compare-using-spec (incompatible spec netnames lit1 lit2 
				   neg-p1 pred1 arglist1 
				   neg-p2 pred2 arglist2 effort &aux
			     (specres (if incompatible
					  (find-symbol "INCOMPATIBLE-LITS" spec)
					  (find-symbol "COMPATIBLE-LITS" spec))))
  (if (null specres)
      nil
      (let ()
	   (setq *current-netname* (car netnames))
      (apply specres (list (car netnames) lit1 lit2 
			   neg-p1 pred1 arglist1 
			   neg-p2 pred2 arglist2 effort)) )
      )
  )

; intersect-specs determines the intersection of specialist lists for the
; two predicates given - i.e. the applicable specialists for resolving
; or factoring with those specialists.

(defun intersect-specs (pred1 pred2 op1 op2 &aux
			      (slist1 (if (symbolp pred1) 
					  (union (getp pred1 'SPECIALISTS)
						 (getp op1 'specialists))))
			      (slist2 (if (symbolp pred2) 
					  (union (getp pred2 'SPECIALISTS)
						 (getp op2 'specialists)))))
  (intersection slist1 slist2)
  )

(defun adjust-subs (r prop1 nprop1 prop2 nprop2 &aux
			 olds1 uses1 olds2 uses2 sub s1 s2 nsubs kill)
  (setf (comparison-info-lit1 r) prop1)
  (setf (comparison-info-lit2 r) prop2)
  (dolist (s (comparison-info-subs1 r))
	  (setq s1 (car s))
	  (setq s2 (cadr s))
	  (setq olds1 (position s1 (prop-args nprop1) :test 'equal))
	  (setq uses1 (subseq (prop-args prop1) olds1 (+ olds1 1)))
	  (setq olds2 (position s2 (prop-args nprop2) :test 'equal))
	  (setq uses2 (subseq (prop-args prop2) olds2 (+ olds2 1)))
	  (setq sub (compare-arglists uses1 uses2))
	  (unless sub (setq kill t))
	  (setq nsubs (append nsubs (car sub)))
	  )
  (setf (comparison-info-subs1 r) nsubs)
  (setq nsubs nil)
  (dolist (s (comparison-info-subs2 r))
	  (setq s1 (car s))
	  (setq s2 (cadr s))
	  (setq olds1 (position s1 (prop-args nprop2) :test 'equal))
	  (setq uses1 (subseq (prop-args prop2) olds1 (+ olds1 1)))
	  (setq olds2 (position s2 (prop-args nprop1) :test 'equal))
	  (setq uses2 (subseq (prop-args prop1) olds2 (+ olds2 1)))
	  (setq sub (compare-arglists uses1 uses2))
	  (unless sub (setq kill t))
	  (setq nsubs (append nsubs (car sub)))
	  )
  (setf (comparison-info-subs2 r) nsubs)
  (if kill
      nil
      r)
  )


(defun compare-with-spec (incompatible netnames neg-p1 prop1 
				  neg-p2 prop2 effort &aux
				  (pr1 (getp prop1 'pred))
				  (pr2 (getp prop2 'pred))
				  (specs (intersection 
					  (specs-for-pred-or-op pr1)
					  (specs-for-pred-or-op pr2)))
				  (arglist1 nil)
				  (arglist2 nil)
				  nprop1 nprop2 
				  result)
  (setq *flattened-lit* nil)
  (unless (null specs)
;	  (setq nprop1 (last1 (flatten prop1)))
;	  (setq nprop2 (last1 (flatten prop2)))
	  (setq nprop1 prop1)
	  (setq nprop2 prop2)
          (when (traced-p 'spec-test)
                (print-line "  Specialists: Trying to determine " 
		     (if incompatible "incompatibility of " "compatibility of ")
		     (getp prop1 'print) " and " (getp prop2 'print))
		(unless (equal (car netnames) '/)
			    (print-it "    in subnet " (car netnames))) )
	  (setq arglist1 (spec-eval-args nprop1))
	  (setq arglist2 (spec-eval-args nprop2))
          (unless (or (null arglist1)
		      (null arglist2))
		  (dolist (s specs)
			  (setq result (compare-using-spec incompatible 
							   s netnames nprop1 nprop2
							   neg-p1 pr1 arglist1
							   neg-p2 pr2 arglist2
							   effort))
			  (unless (null result)
				  ; add other unifiers
				  (unless (and (equal nprop1 prop1)
					       (equal nprop2 prop2))
					  (let (res rr)
					  (dolist (r result)
					   (if (setq rr (adjust-subs r
							   prop1 nprop1
							   prop2 nprop2))
					       (setq res (append res
								 (list rr)))))
					  (setq result res)
					  ))
				  (return result)
				  )
			  )
		  )
	  )
  result)

; resolve-factor strips off the embeding part of the two literals. If both
; refer to the same subnet, that netname is saved, and we continue.
; Each literal is split into negation, predicate and arguments and passed to
; all applicable specialists to tyr to resolve or factor (depending on flag
; resolve).

(defun spec-compare-lits (incompatible lit1 lit2 effort &aux
		          (netprop1 (strip-modals lit1)) 
		          (netname1 (car netprop1))
		          (prop1 (cadr netprop1))
		          (netprop2 (strip-modals lit2)) 
		          (netname2 (car netprop2))
		          (prop2 (cadr netprop2))
			  (depth effort)
			  result 
			  (neg-p1 nil)
			  (neg-p2 nil))
  (when (subsetp netname1 netname2) ; should be good enough for most?
	(setq *current-netnames* netname1)
	(setq *current-embedding* (cadddr netprop1))
	(setq *current-literal* nil)
        (if (null depth)
            (setq depth *specialist-eval-effort*))
        (unless (or (null prop1)
		    (null prop2))
	        ; never get and, or or nec here, but maybe not
		; if 1st has negation pred, save that info 
	        (when (neg-p prop1) (setq neg-p1 t))
  
		; if 2nd has negation pred, save that info 
	        (when (neg-p prop2) (setq neg-p2 t))

		; now call specialist or specialists 
                (setq result (compare-with-spec incompatible netname1 
				        neg-p1 prop1 neg-p2 prop2 depth))
	        )
        )
  (if (equal result 'unknown) nil result))

; incompatible-lits attempts to determine whether two literals are incompatible
; by calling applicalbe specialists.

(defun spec-incompatible-lits (lit1 lit2 &optional effort)
  (spec-compare-lits t lit1 lit2 effort)
  )

; compatible-lits attempts to determine whether two literals are compatible by 
; calling applicalbe specialists.

(defun spec-compatible-lits (lit1 lit2 &optional effort)
  (spec-compare-lits nil lit1 lit2 effort)
  )

(set-display-function 'spec-effort 'spec-effort "Displays specialist effort values")

; spec-effort prints the effort levels used by the
; specialist-interface on entry and evaluation of literals

(defun spec-effort (&rest args)
  (declare (ignore args))
      (print-line "Specialist effort values are:" *nl*
		 #\tab "Entry     " #\tab *specialist-entry-effort* *nl*
		 #\tab "Evaluation" #\tab *specialist-eval-effort*) 
  )

(defun add-interested-party (item spec &aux
				       (int-party (make-interested-party
						   :spec spec
						   :lit *current-literal*
						   :netname *current-netname*)))
  (unless (or (null *current-literal*)
	      (member-if 
	       #'(lambda (i)
		   (and (equal (interested-party-netname int-party)
			       (interested-party-netname i)) 
			(equal (interested-party-lit int-party)
			       (interested-party-lit i)) 
			(equal (interested-party-spec int-party)
			       (interested-party-spec i)) ) 
		   )
	       (getp item 'INTERESTED-PARTY))
	      )
    (if (traced-p 'interested-party)
	(print-line "  Specialist interface: "
		    "Adding to interested party list of " 
		    item *nl* *spaces*
		    "            " spec " " *current-netname* 
		    " " *current-literal* ": "
		    (print-info *current-literal*)))
    (change-property item 'INTERESTED-PARTY
		     (pushnew int-party (get item 'INTERESTED-PARTY) ;; was-getp
			      :test 'equal))
    )
  )

(defun print-int (brief arg)
  (declare (ignore brief))
  (unless (null arg)
	  (print-line "Interested party list for " arg)
	  (print-line #\tab "Specialist" #\tab "Subnet" #\tab "Literal")
	  (mapcar #'(lambda (i)
			    (print-line #\tab (interested-party-spec i)
				        #\tab (interested-party-netname i)
				        #\tab (interested-party-lit i)))
		  (getp arg 'INTERESTED-PARTY))
	  )
  nil)

(defun interested-party (brief &rest args)
  (mapcar #'(lambda (a) (print-int brief a)) args)
  )

(defun incompatible-with-eval (lit1 lit2 eval1 eval2 subs1 subs2 print header
				    &optional (incompatible t))
  (when (traced-p 'spec-test)
        (print-line "    Checking "
		 (if incompatible "incompatiblity" "compatibility")
		 " attempts with evaluations: "
                         (if eval1 eval1 'unknown) " and "
                         (if eval2 eval2 'unknown) )) 
  ;only useful if YES no or NO YES for this system
  (when (if incompatible
	    (or (and (equal eval1 'yes) (equal eval2 'no))
		(and (equal eval1 'no) (equal eval2 'yes)))
	    (or (and (equal eval1 'yes) (equal eval2 'yes))
		(and (equal eval1 'no) (equal eval2 'no)))) ;?
	      (when print
		    (print-line header "Found " (getp lit1 'print)
				   (if incompatible " incompatible "
				       " compatible ")
				   "with " (getp lit2 'print))
		    (when (or subs1 subs2)
			  (print-line "      with substitutions: "
					 (print-subs (append subs1 subs2)))))
	      (list (make-comparison-info :lit1 lit1 :lit2 lit2 
				      :subs1 subs1 :subs2 subs2)))
  )

;; the *changed-concepts* list contains the names of concepts which have
; had information changed inside them as a result of the current assertion
; (for example, propagation). The interested party lists of these are
; obtained after the current assertion is finished, and reasserted.

(defvar *changed-concepts* nil " Keeps track of items changed inside specs")

; spec-changed-concept is available so that a specialist can call it to
; makr a concept as changed.

(defun spec-changed-concept (concept)
  (pushnew concept *changed-concepts*)
  )

(defun make-spec-inference (formula)
  (if (equal *current-netname* '/)
      formula
      ; else add modal embedding
      (subst formula 'embedded *current-embedding* :test 'equal)
      )
  )

(defun remove-episodic (list)
  (cond ((null list) nil)
	((atom list) list)
	((member (cadr list) '(** * @)) (remove-episodic (car list)))
	(t list))
  )

(defun check-subnet-equal (&aux formulas args form)
  ; grab formulas based on top level
  ; each must be checked to make sure embedding same
  ; want to find subnet formulas and duplicate them with new equivalence
  (when (and (equal (getp *current-literal* 'wff-type) 'prefix)
	     (equal (getp *current-literal* 'pred) 'equal)
	     (setq args (getp *current-literal* 'arguments))
	     (symbolp (car args)) (symbolp (cadr args))
	     )
    (setq form
      (remove-episodic
       (if (equal *current-embedding* 'embedded)
	   (print-info *current-literal*)
	 *current-embedding*)))
    (when (and (listp form) (symbolp (car form)))
      (setq formulas (ret (car form) 'tp.mental-attitude nil t t))
      )
    (dolist (f formulas) (check-subnet-formula f))
    ))


; uses same extraction techniques as strip-prop - if one changes so should
; the other
(defun check-subnet-formula (whole-formula &optional 
					   (formula whole-formula)
					   (subnets '(/))
					   (embedding 'embedded))
  ; try every combo of equivalences throughout entire formula, and enter
  (case (wff-type formula)
	(prefix 
	 (if (subnet-pred (getp formula 'pred))
	     (equiv-subnet-formulas whole-formula formula subnets embedding)
	     (if (equal (car subnets) '/) 
		 nil
		 ;enter formula
		 (progn
		       (setq *current-embedding* embedding)
		       (spec-enter-formula subnets formula (getp formula 'pred))
		       ))
	     ))
	(episodic
	 (unless (equal (getp formula 'pred) '_)
		 (check-subnet-formula whole-formula 
	            (car (getp formula 'arguments)) subnets
		    (subst '* '** (subst embedding 
			 (print-info (car (getp formula 'arguments))) 
			 (print-info formula) :test 'equal)) )
		 ))
	)
  )

; uses same extraction techniques as modal-prop - if one changes so should
; the other
(defun equiv-subnet-formulas (whole-formula formula netnames embedding &aux 
					    args pred wff diff)
  (setq args (getp formula 'arguments))
  (setq pred (getp formula 'pred))
  ; last argument indicates formula
  (setq wff (last1 args))
  (setq args (remove wff args))

  (cond ((null wff) nil)
	((equal (getp wff 'entity-type) 'term)
	 (case (term-type wff)
	       (constant
		(cond ((getp wff 'print) (setq wff (getp wff 'print)))
		      ((equal (find-sort wff) 'propos)
		      ; find content, use that
		      (setq diff t)
		      (let ((ws (get-class (list wff '_ 'object))))
			   (if ws 
			       (setq wff (car (getp (select-one ws) 'arguments)))
			       (setq wff nil))
		      )) ))
	       (modified
		(if (equal (getp wff 'operator) 'that)
		    (setq wff (car (getp wff 'arguments)))
		    ; what else can be done here?
		    ))
	       (wff nil)
	       (variable (setq wff nil))
	       (otherwise (setq wff nil))))
	((equal (getp wff 'entity-type) 'wff) nil)
	(t (setq wff nil))
	)
  (when wff
	(unless diff
	(setq embedding (subst embedding (print-info wff) (print-info formula)
				:test 'equal)) )
	(build-new-formula whole-formula netnames embedding wff args pred)
	)
  )

(defun build-new-formula (whole-formula netnames embedding formula args pred)
  ; make list of applicable subnet names
  (setq netnames (make-all-netnames netnames args pred))
  (check-subnet-formula whole-formula formula netnames embedding) 
  )

(defun make-netnames (netname args arglists pred &aux
			      nlist)
  (cond ((null arglists)
	 (setq nlist (list (build-symbol (append (list netname) 
			      (if (equal netname '/) nil (list '/))
			      args
			      (list '-) (list pred)) 'epilog))))
	(t (dolist (a (car arglists))
		   (setq nlist (append nlist 
				       (make-netnames netname 
						      (append args (list a))
						      (cdr arglists)
						      pred)))
		   ))
	)
  nlist)

;; Extract specialist-specific representation from an environment.
;; env is an environment; spec is a symbol that identifies the
;; specialist; fn, if provided, is a function of no arguments that, if
;; the given specialist has not yet registered any information in the
;; environment, will be called to initialize that specialist's data
;; structures.

(defun retrieve-specialist-info (env spec-name &optional fn)
  (let* ((alist (environment-specialist-info env))
	 (result (assoc spec-name alist)))
    (if result
	(cdr result)
      (if fn
	  (let ((init (funcall fn)))
	    (change-field 'environment-specialist-info
			  env
			  (cons (cons spec-name init)
				alist))
	    init)))))
