;       Last change    August 1993
(in-package epilog)

;; Subgoal routines
;; These routines are responsible for creating and loading new subgoals,
;; given information from the matching of the two wffs.

;; When two non-conditional wffs are being used to create a subgoal, the
;; mathcing literal is removed from the second (accessed) wff, and 
;; replaced by the residue (or removed if there is no residue) from the first
;; (set-of-support) wff. Then the two resulting wffs are combined with an or.
;;     eg.  (A x ((not x girl) or (not x pretty)))   matches  (lrrh girl)
;;       literal (not x girl)                              (lrrh girl)
;;       subs    lrrh/x
;;  new subgoal:  (not lrrh pretty)

;; If one is conditional, and the other isn't, the same procedure is followed,
;; using the consequent of the conditional and the other wff. The negation of
;; the antecedent of the conditional is also incorporated into the subgoal.
;;     eg. (A x (x girl) (x pretty))  matches 
;;                                 ((not lrrh pretty) or (not lrrh happy))
;;       literal (x pretty)                         (not lrrh pretty)
;;       subs    lrrh/x
;;   new subgoal:  ((not lrrh girl) or (not lrrh happy))

;; If both are conditional, only their consequents have been compared so far.
;; First the antecedents are compared for compatibility - if they are not
;; compatible, we stop here. If they are, we create a new subgoal,
;; consisting of a combination of the consequents with their matching
;; literals removed (or replaced by a residue), and the negation of the
;; combination (and) of the two antecedents, with one of the matching
;; literals removed.
;;      eg. (A x (x girl) (x pretty))   matches (A x (x girl) (not x pretty))
;;    (proof attempt of (E x (x girl) (x pretty)) )
;;           literal (x pretty)                  (not x pretty)
;;           subs    x/x
;;   new subgoal:   (A x (not x girl))

(traceable 'subgoal "Traces the addition and splitting of subgoals")

(tweakable '*quantified-difficulty* 20 
  "Difficulty value for non-conditional quantified subgoals"
  '(float * *))
(tweakable '*split-difficulty* 10
  "Difficulty value for simple splits of subgoals" '(float * *))
(tweakable '*conditional-difficulty* 30
  "Difficulty value for conditional subgoals where antecedent is assumed"
  '(float * *))
(tweakable '*assume-difficulty* 30
  "Difficulty value for subgoals requiring assumptions, other than conditionals"
  '(float * *))


; Splitting

(defun add-subgoal (parent subgoal &key subs proof-prop assumptions (prob 1)
					(c-prob 1) non-circular 
					(difficulty 0) (depth 0) wffs rules disallowed-rules
		    &aux pari oi usei sca sol pred s-item)
  (when (and parent (null proof-prop)) 
	(setq proof-prop (subgoal-proof-prop parent)))
  (when (and (equal depth 0) (null parent) (equal proof-prop 'PROOF)) 
	(setq *wh-pattern* nil)
	(setq *e-pattern* nil))
  (unless (or subs (not parent)) (setq subs (subgoal-subs parent)))
  (setq subs (remove-if-not #'(lambda (a) (member (car a) (append *wh-pattern* *e-pattern*))) subs))
  (setq subgoal (normalize (print-info subgoal) '-))
  (when (equal proof-prop 'PROOF)
	(setq subs (combine-subs subs *normed-subs*)))
  (when (and (equal depth 0) (null parent) (equal proof-prop 'PROOF)) 
	(setq *normed-question* subgoal)
        (set-wh-pattern subgoal)
        (if *save-results* (set-e-pattern subgoal)))
  (when (and parent (not (equal subgoal 'YES))
	     (member-if #'(lambda (a) (equal (subgoal-subgoal a) subgoal))
			(subgoal-subgoals parent)))
	(when (traced-p 'qa-success)
	      (print-line (getp subgoal 'print) " already a subgoal of " 
		    (getp (subgoal-subgoal parent) 'print)))
	(return-from add-subgoal nil))
  (when parent
	(setq pari (subgoal-interest parent))
	(setq oi (subgoal-inherit-interest parent))
	(if (> pari oi) (setq usei pari) (setq usei oi)))
  (setq s-item (make-subgoal :subgoal subgoal
			     :proof-prop proof-prop
			     :prob prob
			     :subs subs
			     :combo 'or
			     :c-prob (calc-prob-list c-prob)
			     :difficulty difficulty
			     :assumptions assumptions
			     :wffs wffs
			     :previous (if parent
					   (append (subgoal-previous parent)
						   (list subgoal))
					   (list subgoal))
			     :parent parent
			     :interest (wff-interest subgoal)
			     :inherit-interest 
			      (if parent
			      (inherit-calc usei
			       (subgoal-c-prob parent) (calc-prob-list c-prob))
				  *initial-charge*)
			     :depth depth
			     :rules rules
			     :disallowed-rules disallowed-rules
			     :complexity (wff-complexity subgoal)))
  (when parent
	(setf (subgoal-subgoals parent) (append (subgoal-subgoals parent)
						(list s-item))))
  (if (equal subgoal 'YES)
      (return-from add-subgoal 
	(finish-subgoal s-item (list subgoal wffs prob) subs depth)))
  (when (and (not non-circular) parent 
	     (member subgoal (subgoal-previous parent)))
	(when (traced-p 'qa-success)
	      (print-line "Circular subgoal " (getp subgoal 'print) 
			  " - stopping this chain"))
	(if (equal (subgoal-combo parent) 'and) ;bad news
	    (setf (subgoal-answer s-item) '(NO nil 1)))
	(return-from add-subgoal nil))
  (when (and (getp subgoal 'subgoals) 
	     (if non-circular
		 (member-if #'(lambda (a) 
		  (null (set-exclusive-or (subgoal-assumptions a) assumptions)))
			    (getp subgoal 'subgoals))
		 t)
	     (member-if #'(lambda (a) (equal (subgoal-proof-prop a) proof-prop))
			(getp subgoal 'subgoals)))
	(when (traced-p 'qa-success)
	      (print-line "Subgoal " (getp subgoal 'print) 
			  " rederived on another chain " 
			  (subgoal-proof-prop (car (getp subgoal 'subgoals)))
			  " - stopping this chain"))
	(if (and parent (equal (subgoal-combo parent) 'and)) ;bad news
	    (setf (subgoal-answer s-item) '(NO nil 1)))
	(return-from add-subgoal nil))
  (when (> depth *qa-depth*)
	(when (traced-p 'qa-test)
	      (print-line "Maximum depth reached for this line of questioning"
			  " - stopping it"))
	(if (and parent (equal (subgoal-combo parent) 'and)) ;bad news
	    (setf (subgoal-answer s-item) '(NO nil 1)))
	(return-from add-subgoal nil))
  (change-property subgoal 'subgoals (list s-item))
  (pushnew subgoal *wff-subgoals*)
  (when (traced-p 'subgoal)
	(print-line "Adding " (subgoal-proof-prop s-item) " subgoal to prove: " 
		    (getp subgoal 'print) ))

  ; make any necessary assumptions
  ; build new subgoals, carrying parents assumptions and new ones

  (case (getp subgoal 'wff-type)
	(quantified 
	 (cond ((and (getp subgoal 'conditional) (not (neg-p subgoal)))
		(if (< *question-effort* 3) (return-from add-subgoal nil))
		(let ((ante (list 'E (getp subgoal 'variable) 
				  (getp (getp subgoal 'restrict) 'print)))
		      (conse (getp subgoal 'main))
		      (cprob 
			 (cond ((getp (getp subgoal 'quantifier) 'prob)) (t 1)))
		      )
		     (conditional-subgoal s-item ante conse cprob)
		     ))
	       (t ; non-conditionals
		  (setf (subgoal-difficulty s-item) 
			    (+ (subgoal-difficulty s-item) 
			       *quantified-difficulty*));)
		  ; add to agenda
		  (add-access-actions s-item))
	       ))
	(logical
	 (cond ((get subgoal 'equivalence)
		(if (< *question-effort* 3) (return-from add-subgoal nil))
		(let* ((args (getp subgoal 'arguments))
		       (ante (car args))
		       (conse (cadr args))
		       (cprob 1)
		       )
		      (setf (subgoal-combo s-item) 'and)
		      (conditional-subgoal s-item (getp ante 'print) conse cprob)
		      ; if quantification involved in ante, should transfer
		      ; to conse before this step
		      (setq sca (switch-quant ante conse))
		      (conditional-subgoal s-item (car sca) (cdr sca) cprob
					   )
		      )
		)
	       ((getp subgoal 'conditional)
		(if (< *question-effort* 3) (return-from add-subgoal nil))
		(let* ((ante (getp (car (getp subgoal 'arguments)) 'print))
		      (conse (cadr (getp subgoal 'arguments)) )
		      (pr (getp subgoal 'pred))
		      (cprob (cond ((numberp pr))
				   ((and (listp pr) (numberp (car pr))
					 (<= (car pr) 1.0)) (car pr))
				   (t 1)))
		     ; (if (numberp (getp subgoal 'pred)) (getp subgoal pred) 1)
		      )
		     (conditional-subgoal s-item ante conse cprob)
		     ))
	       ; check to see if there are any quantifiers involved - if so,
	       ; the subgoals may not be independent
	       ((not (independent-subgoals (getp subgoal 'arguments)))
		(setf (subgoal-difficulty s-item)
		      (+ (subgoal-difficulty s-item) *quantified-difficulty*))
		(add-access-actions s-item))
	       ; ands, and ors where assumptions are not allowed
	       ((or (equal (setq pred (getp subgoal 'pred)) 'and)
		    (and (< *question-effort* 3) (equal pred 'or)))
		; make a new subgoal for each argument, combine with and
		(when (traced-p 'subgoal)
		      (print-line "Splitting subgoal " (getp subgoal 'print) 
				  " into " pred "ed subgoals"))
		(setf (subgoal-combo s-item) pred)
		(mapcar #'(lambda (a) (add-subgoal s-item a
						   :difficulty 
						   *split-difficulty* ;10 ;0
						   :subs subs
						   :prob prob ;1
						   :c-prob (calc-prob-list prob) ;prob
						   :rules nil ;rules
						   :disallowed-rules 
						     nil ;disallowed-rules
						   :wffs nil ;wffs
						   :depth (+ depth 1)
						   :assumptions assumptions))
			(getp subgoal 'arguments))
		)
	       (t ;or - assume negation of all except 1 argument, prove other
		  (let* ((args (getp subgoal 'arguments))
			 (prove (car args))
			 (neg (cdr args))
			 assump newass)
			(setq assump (print-info (car neg)))
			(when (cdr neg)
			      (setq assump (list assump 'or))
			      (setq assump (append assump 
			        (mapcar 'print-info (cdr neg)))) )
			(setq assump (negate assump))
			(when (traced-p 'subgoal)
			      (print-line "Assuming " assump " " proof-prop))
			(setq *skol-subs* nil)
			(setq newass (remove 'YES (stuff-it assump :proof-prop proof-prop
					       :assumption t)))
			(setq newass (remove 'INCONSISTANT newass))
			(setq prove 
			      (normalize (substitute-prop prove *skol-subs*)))
			(if (member 'NO newass) 
			    (return-from add-subgoal 
			      (finish-subgoal s-item (list 'YES
							   (subgoal-wffs s-item) prob)
					      subs depth)))
			(when (traced-p 'qa-eval)
			      (print-line "  Verifying: " (print-info prove)))
			(setq sol (verify prove proof-prop nil nil assumptions))
			(when (and (traced-p 'qa-eval) sol)
			      (print-it "     --> " (print-info (car sol))))
			(when sol 
			      (setq prove (car sol))
			      (setq prob (combine-probs prob (caddr sol)))
			      (setq wffs (union wffs (cadr sol))))
			(reset-skol-subs *skol-subs*)
			(if (and sol (member (car sol) '(YES NO inconsistent)))
			    (finish-subgoal s-item sol subs depth)
			    (add-subgoal s-item prove
				     :difficulty *assume-difficulty* ;20 ;0
				     :subs subs
				     :depth (+ depth 1)
				     :prob 1
				     :c-prob prob
				     :rules rules
				     :disallowed-rules disallowed-rules
				     :wffs nil ;wffs
				     :assumptions (append assumptions newass)))
			(if (not *goal-forward*) ;add others of disjunction also
			    (mapcar #'(lambda (a) (add-subgoal s-item a
					           :difficulty *split-difficulty* ;10 ;0
						   :subs subs
						   :c-prob prob
						   :prob 1
						   :rules rules
						   :disallowed-rules 
						     disallowed-rules
						   :wffs nil ;wffs
						   :depth (+ depth 1)
						   :assumptions assumptions))
				    (cdr args)))
			))
	       ))

	; others cannot be split
	(otherwise
	 ; add to agenda - add access actions for it
	 (add-access-actions s-item))
	)
  s-item)

(defun independent-subgoals (slist)
  (if (member-if #'(lambda (a) (getp a 'quantifier)) slist)
      ; check for unquantified variables in slist
      (not (member-if 'look-for-unquant-vars slist))
      t)
  )

(defun look-for-unquant-vars (formula &optional quant-vars vars)
  (when (and formula (symbolp formula))
	(case (getp formula 'wff-type)
	      (quantified
	       (setq vars (append quant-vars (list (getp formula 'variable))))
	       (or (look-for-unquant-vars (getp formula 'restrict) vars)
		   (look-for-unquant-vars (getp formula 'main) vars)) )
	      (modified
	       (look-for-unquant-vars (car (getp formula 'arguments)) quant-vars))
	      (logical
	       (eval (append (list 'or) (mapcar #'(lambda (a) (look-for-unquant-vars a quant-vars)) (getp formula 'arguments)))) )
	      ((prefix causal episodic)
	       (eval (append (list 'or) (mapcar #'(lambda (a) (look-for-unquant-vars a quant-vars)) (getp formula 'arguments)))) )
	      (otherwise
	       (if (and (epi-variable formula) (not (member formula quant-vars)))
		   t
		   nil))
	      ))
  )


(defun select-negs (args &aux quant q neg prove rem)
  (mapcar #'(lambda (a) 
		    (if (and (setq q (getp a 'quantifier)) 
			     (equal q 'A))
			(setq quant (append quant (list a))))) args)
  (setq rem (set-difference args quant))
  (cond ((null rem)
	 (setq prove (car quant))
	 (setq neg (cdr quant)))
	(t (setq prove (car rem))
	   (setq neg (append quant (cdr rem))) ))
  (cons neg prove)
  )

(defun reset-skol-subs (subs)
  (dolist (s subs)
	  (remprop (car s) 'skol-const)))

(defun conditional-subgoal (s-item ante conse cprob &aux newass subs sol
				   (wffs (subgoal-wffs s-item))
				   (proof-prop (subgoal-proof-prop s-item))
				   (prob (subgoal-prob s-item)))
  ;; I don't know if cprob is supposed to be ignored, but it is, and I
  ;; don't know what to do with it. -ANK
  (declare (ignore cprob))
  ; To Prove: assume antecedent, prove consequent
  ; antecedent is in list form, conse is in normalized form
  (when (traced-p 'subgoal) (print-line "Assuming " ante " " proof-prop))
  (setq *skol-subs* nil)
  (setq newass (remove 'YES 
		       (stuff-it ante :proof-prop proof-prop :assumption t )))
  (setq newass (remove 'INCONSISTANT newass))
  (if (member 'NO newass) 
      (return-from conditional-subgoal 
		   (finish-subgoal s-item (list 'YES (subgoal-wffs s-item) prob)
				 (subgoal-subs s-item) (subgoal-depth s-item))))
  (setq conse (normalize (substitute-prop conse *skol-subs*)))
  (when (equal (subgoal-proof-prop s-item) 'PROOF)
	(setq subs (combine-subs (subgoal-subs s-item) *normed-subs*)))
  (when (traced-p 'qa-eval)
	(print-line "  Verifying: " (print-info conse)))
  ; ** changed to include the assumed wff earlier in the verification
  (setq newass (append (subgoal-assumptions s-item) newass)) ; **
  (setq sol (verify conse (subgoal-proof-prop s-item) nil nil newass))
		    ; ** (( (subgoal-assumptions s-item)))
  (when (and (traced-p 'qa-eval) sol)
	(print-it "     --> " (print-info (car sol))))
  (when sol 
	(setq conse (car sol))
	(setq prob (combine-probs prob (caddr sol)))
	(setq wffs (union wffs (cadr sol))))
  (reset-skol-subs *skol-subs*)
  (if (and sol  (member (car sol) '(YES NO inconsistent)))
      (return-from conditional-subgoal
		       (finish-subgoal s-item sol subs (subgoal-depth s-item)))
  ; how should the prob should be used?
  (add-subgoal s-item conse 
	       :subs subs ; (subgoal-subs s-item)
	       :rules (subgoal-rules s-item)
	       :disallowed-rules (subgoal-disallowed-rules s-item)
	       :wffs nil ;wffs 
	       :assumptions newass ; (append (subgoal-assumptions s-item) newass)
	       :c-prob  prob 
	       :prob 1
	       :depth (+ (subgoal-depth s-item) 1)
	       :difficulty *conditional-difficulty* ;10 ;0
	       ) )
  )

(defun finish-subgoal (subgoal answer subs &optional (depth 0) &aux)
  (if (done-subgoal subgoal answer subs depth)
      (if (and *qa-mode* 
	       (or (null *qa-agenda*) *key-verify* *exists-q* *insertion*
		   (if *wh-pattern* 
		       (and *number-answers* 
			    (>= (length *solutions*) *number-answers*)
			    (let () (print-line "Stopping after " 
					    *number-answers* " answers") 
				 t))
		       (or (= (calc-prob-list (caddr (combine-answers))) 1)
			   (if (>= *qa-number-iterations* *minimum-effort*)
			       t
			       (let ()
				(setq *continue-til-minimum* *minimum-effort*)
				nil)) )
		       )))
	  (throw 'question-catcher t)) )
  ; else move everything down on agenda
  )

(defun reset-agenda (subgoal)
  (dolist (a (subgoal-access-items subgoal))
	  (setq *qa-agenda* (remove a *qa-agenda*))
	  (setf (agenda-item-position a)
		(+ (agenda-item-position a) *finished-weight*))
	  (add-agenda a nil))
  (dolist (a (subgoal-agenda subgoal))
	  (setq *qa-agenda* (remove a *qa-agenda*))
	  (setf (agenda-item-position a)
		(+ (agenda-item-position a) *finished-weight*))
	  (add-agenda a nil))
  (dolist (s (subgoal-subgoals subgoal))
	  (unless (subgoal-answer s) (reset-agenda s)))
  )

; perhaps have answer be the full yes wffs prob answer instead of just yes
(defun done-subgoal (subgoal answer subs &optional (depth 0) &aux parent 
								  (apart (car answer))
								  (wpart (cadr answer))
								  (ppart (caddr answer)))
  ; take answer up to parent and see how to combine
  (when (and (traced-p 'subgoal) (not (equal (print-info (subgoal-subgoal subgoal)) 'YES)))
	(print-line "Answer " (car answer) " for " (print-info (subgoal-subgoal subgoal))))
  (setf (subgoal-answer subgoal) answer)
  (cond ((null (setq parent (subgoal-parent subgoal)))
	 ; top level - have an actual solution
	 (when (equal (car answer) 'YES)
	       (when (solution 
		      (list (if (equal (subgoal-proof-prop subgoal) 'PROOF)
				 'YES 'NO)
			    (union (cadr answer) 
				   (subgoal-wffs subgoal))
			    (caddr answer)) subs depth)
		     (setf (subgoal-useful subgoal) t)
		     t)) )
	(t (unless *wh-pattern* (reset-agenda subgoal))
	   (cond 
	((equal (length (subgoal-subgoals parent)) 1)
	 (setf (subgoal-useful subgoal) t)
	 (done-subgoal parent (list apart (union (subgoal-wffs subgoal) wpart) 
				    (combine-probs (subgoal-prob parent)
						   ppart)) subs depth))
	((and (equal (car answer) 'YES) (equal (subgoal-combo parent) 'or))
	 (setf (subgoal-useful subgoal) t)
	 (done-subgoal parent (list apart (union (subgoal-wffs subgoal) wpart) 
				    (combine-probs (subgoal-prob parent)
						   ppart)) subs depth))
	((and (equal (car answer) 'NO) (equal (subgoal-combo parent) 'and))
	 (setf (subgoal-useful subgoal) t)
	 (done-subgoal parent (list apart (union (subgoal-wffs subgoal) wpart) 
				    (combine-probs (subgoal-prob parent)
						   ppart)) subs depth))
	(t ; if all answers are in now, check to see if finished
	   (let ((ans (mapcar 'subgoal-answer (subgoal-subgoals parent)))
		 (pr 1)
		 (wf wpart))
		(setq pr (subgoal-prob parent))
		(dolist (a ans)
			(setq pr (combine-probs pr (caddr a)))
			(setq wf (union wf (cadr a))))
		(cond ((and (equal (subgoal-combo parent) 'and)
			    (not (member-if-not #'(lambda (a) (equal a 'YES))
						(mapcar 'car ans))))
		       (setf (subgoal-useful subgoal) t)
		       (dolist (s (subgoal-subgoals parent))
			       (setf (subgoal-useful s) t))
		       (done-subgoal parent (list 'YES wf pr) subs depth)
		       )
		      ((and (equal (subgoal-combo parent) 'or)
			    (not (member-if-not #'(lambda (a) (equal a 'NO))
						ans)))
		       (setf (subgoal-useful subgoal) t)
		       (dolist (s (subgoal-subgoals parent))
			       (setf (subgoal-useful s) t))
		       (done-subgoal parent (list 'NO wf pr) subs depth)
		       )
		      ))
	   ))))
  )


(defvar *unique-var-num* 1)

(defun distinct-var (v &aux vname)
  (setq vname (build-symbol (list 'VAR *unique-var-num*) 'epilog))
  (setq *unique-var-num* (+ *unique-var-num* 1))
  (make-variable vname)
  (setp vname 'SORT (getp v 'SORT))
  (setp vname 'MAIN-SORT (getp v 'MAIN-SORT))
  (setp vname 'MATCHABLE (getp v 'MATCHABLE))
  vname)

(defun fix-mult-vars (wff subs1 subs2 repl2 new2 &aux new1 subs2r newsubs)
  (dolist (i (mapcar 'car (getp wff 'epilib::types)))
          (setq subs2r (push (list i (distinct-var i) t) subs2r)))
  (dolist (s subs2)
	  (setq newsubs (append newsubs (list (list 
		(normalize-term (substitute-term (car s) subs2r nil 1 nil) '+) 
		(cadr s) 
		(caddr s))))) )
  (dolist (s subs1)
	  (setq new1 (append new1 (list (list (car s) 
		(normalize-term (substitute-term (cadr s) subs2r nil 1 nil) '+) 
		(caddr s))))) )
  (setq wff (normalize-wff (substitute-prop wff subs2r) '+))
  (setq repl2 (normalize-wff (substitute-prop repl2 subs2r) '+))
  (setq new2 (normalize-wff (substitute-prop new2 subs2r) '+))
  (list wff new1 newsubs repl2 new2)
  )

(defun try-exist-subgoal (s-item subs2 phi kn-wff proof-prop  &optional (newsub t)
			       &aux fire newass)
  ;(setq *key-threshold* 1.0)
  (setq subs2 
	(remove 'nil (mapcar #'(lambda (s) (if (epi-variable (cadr s)) nil s)) subs2)))
  (cond (subs2
	 (setq fire (substitute-prop kn-wff subs2))
	 (when (traced-p 'subgoal)
	       (print-line "Instantiating " (print-info kn-wff) 
			   " with substitutions " (print-subs subs2)))
	 (setq newass (remove 'YES 
		       (stuff-it fire :proof-prop proof-prop :assumption t
				 :assumptions (subgoal-assumptions s-item))))
	 )
	(t 
	 (when (getp kn-wff 'assumption) (return-from try-exist-subgoal nil))
	 (when (traced-p 'subgoal)
	       (print-line "Looking for inferences with " (print-info kn-wff) ))
	       
	 (unless (and (null (subgoal-assumptions s-item))
		      (getp kn-wff 'fired))
	 (setq newass (remove kn-wff (remove 'YES 
		  (reassert kn-wff :assumptions (subgoal-assumptions s-item)
			    :proof-prop proof-prop :qa t))) )
	 (when (null (subgoal-assumptions s-item))
	       (change-property kn-wff 'fired t)) )
	 ) )
  ;(setq *key-threshold* kt)
  (if (or (member 'NO newass) (member 'inconsistent newass))
      (return-from try-exist-subgoal nil))
  (when newass
	(if newsub
  (add-subgoal s-item phi 
	       :subs (subgoal-subs s-item) ; + subs2?
	       :rules (append (subgoal-rules s-item) (list (list kn-wff)))
	       :disallowed-rules (append (subgoal-disallowed-rules s-item)
					 (list kn-wff))
	       :wffs (list kn-wff) 
	       :non-circular t
	       :assumptions (append (subgoal-assumptions s-item) newass)
	       :c-prob  (subgoal-c-prob s-item) 
	       :prob (subgoal-prob s-item) ;1
	       :depth (+ (subgoal-depth s-item) 1)
	       :difficulty 0 
	       ) 
	    newass)
	    ) 
  )

(defun exist-match (s &aux
		      (v (cadr s)))
  (and (epi-variable v) (not (getp v 'matchable)) (not (getp v 'skol-const))
       (not (getp v 'lambda))))

(defun complex-kn-phi (kn-wff phi subs1 subs2) 
  (cond ( (and (or (getp kn-wff 'conditional)
	   (and (equal (getp kn-wff 'pred) 'OR)
		(member-if #'(lambda (a) (equal (getp a 'wff-type) 'quantified))
			   (getp kn-wff 'arguments))))
       (member-if 'exist-match subs1)
       (> (list-length (append (getp phi '+keys) (getp phi '+other)
			       (getp phi '-keys) (getp phi '-other))) 1)
       )
	  t)
	((null subs2) nil)
	(t nil)
	;(t (print-line "instant for " (getp kn-wff 'print) " " subs1 " and "
	;	       *nl* (getp phi 'print) subs2)
	;   (setq ans (read-line))
	;   (equal ans "YES"))
	)
  )

(defun subgoal (item &aux 
		     (s-item (agenda-item-subgoal-parent item))
		     (assumptions (subgoal-assumptions s-item))
		     (proof-prop (subgoal-proof-prop s-item))
		     (phi (subgoal-subgoal s-item))
		     (subs (subgoal-subs s-item))
		     (prob (agenda-item-prob item))
		     (gparents nil) 
		     (depth (subgoal-depth s-item))
		     (subgoal-info (agenda-item-subgoal-action item))
		     (success (subgoal-info-comparison subgoal-info))
		     (key (subgoal-info-key subgoal-info))
		     (keycon (subgoal-info-kn-wff-context subgoal-info))
		     (key2 (subgoal-info-kn-wff-key subgoal-info))
		     (kn-wff (subgoal-info-kn-wff subgoal-info))
		     (pr-wff kn-wff)
		     s use-lit temp pr np
		     subgoal aphi 
		     newwff usephi parents repl2 new2
		     subs1 subs2 
		     (useprob prob))
  (when (setq s success)
	(setf (subgoal-agenda s-item) (remove item (subgoal-agenda s-item) :test 'equal))
	(cond ((equal key (comparison-info-lit1 s))
	       (setq subs1 (comparison-info-subs1 s))
	       (setq subs2 (comparison-info-subs2 s))
	       (setq repl2 (comparison-info-lit2 s)) )
	      (t (setq subs1 (comparison-info-subs2 s))
		 (setq subs2 (comparison-info-subs1 s))
		 (setq repl2 (comparison-info-lit1 s)) ) )

	; if any subs are for existential variables in a rule (embedded)
	; escape and try different method
	(if (and (not (comparison-info-residue1 s)) 
		 (not (comparison-info-residue2 s))
		 ; this test should be more general and include
		 ; embedded existentials in disjunctions too - but what
		 ; to do with them after?
		 (complex-kn-phi kn-wff phi subs1 subs2)
		 )
	    (return-from subgoal (try-exist-subgoal s-item subs2 phi kn-wff 
						    proof-prop ))
	    )

          (setq new2 (normalize (cond ((comparison-info-residue2 s))
			   ((comparison-info-residue1 s))
			   ((equal keycon '-) 'NO)
                           (t 'YES)) '+ nil nil)   )
	  ; if the formula is a () number () formula, change to not () or ()
	  ; first ?
  	   (if (setq np (cond ((numberp (setq pr (getp new2 'pred))))
			      ((and (listp pr) (numberp (car pr))
				    (<= (car pr) 1.0)) (car pr))))
		    (setq prob (combine-probs prob np)))
	  (setq new2 (normalize-wff (negate new2) '+))
          (when (and (getp kn-wff 'epilib::types) (getp phi 'epilib::types))
                (setq temp (fix-mult-vars kn-wff subs1 subs2 repl2 new2))
                (setq kn-wff (car temp))
		(setq subs1 (cadr temp))
                (setq subs2 (caddr temp))
                (setq repl2 (cadddr temp))
                (setq new2 (car (cddddr temp)))
                )
          (when (equal proof-prop 'PROOF)
		(setq subs (combine-subs subs subs1 t)))
          (setq subs2 (append subs2 (list (list repl2 new2))))
 
          (setq useprob prob)
          (setq newwff (substitute-prop kn-wff subs2))
	  (when (and (listp newwff) (numberp (cadr newwff)))
		(setq useprob (combine-probs useprob (cadr newwff))))
          (setq newwff (negate newwff))
          (setq useprob (combine-probs useprob *subst-prob*))
          (setq parents gparents)
	  (setq parents (union gparents (list pr-wff)))
 
	  (setq depth (+ depth 1))
	  (setq use-lit key)
          (setq subs1 (append subs1 (list (list use-lit newwff))))
          (setq usephi (substitute-prop phi subs1))
          (setq useprob (combine-probs useprob *subst-prob*))
          (setq subgoal (normalize usephi '+ nil nil)) 
          (when (equal proof-prop 'PROOF)
		(setq subs (combine-subs subs *normed-subs*)))
	  (when (traced-p 'qa-eval)
			(print-line "  Verifying: " (print-info subgoal)))
          (setq aphi (verify subgoal proof-prop nil nil assumptions))
	  (when (and (traced-p 'qa-eval) aphi)
		(print-it "     --> " (print-info (car aphi))))

	  (if aphi (setq subgoal (car aphi)))
          (when aphi
                (setq useprob (combine-probs useprob 
                                             (if (and (numberp (caddr aphi))
                                                      (cadr aphi))
                                                 (cadr aphi)
                                                 (caddr aphi))))
                (setq parents (union parents (cadr aphi)))
                (unless (member subgoal '(YES NO inconsistent))      
			(setq subgoal (normalize-wff (getp subgoal 'print) '+))
			(when (equal proof-prop 'PROOF)
			      (setq subs (combine-subs subs *normed-subs*)))) 
		)
	(if (equal subgoal 'inconsistent) (return-from subgoal nil))
	(when (and subgoal (traced-p 'qa-success))
	      (print-line "Subgoal" " using key " (print-info key) " in the "
			  (if (equal (subgoal-proof-prop s-item) 'PROOF) 
			      "proof " "disproof ")
			  "subgoal " (print-info phi))
	      (print-line "     and " (print-info key2) " in " 
			  pr-wff ": " (print-info pr-wff))
	      (unless (traced-p 'qa-test)
		      (print-line "       with substitutions: ")
		      (print-trace-subs (comparison-info-subs1 success))
		      (print-it " in " 
				(getp (comparison-info-lit1 success) 'print))
		      (print-line "                      and  ")
		      (print-trace-subs (comparison-info-subs2 success))
		      (print-it " in " 
				(getp (comparison-info-lit2 success) 'print))
		      )
	      (print-line "    yielding")
	      (if (and subgoal (not (equal subgoal 'NO)))
		  (if (equal subgoal 'YES) (print-it " ... " subgoal
						     "       depth " depth)
		      (print-it " ..." *nl* *spaces* (print-info subgoal)
				"       depth " depth))
		  (print-it " the null clause"))
	      )

	(unless (equal subgoal 'NO)
	      (add-subgoal s-item subgoal 
			   :prob useprob
			   :c-prob useprob
			   :subs subs
			   :wffs parents
			   :rules (agenda-item-rules item)
			   :disallowed-rules (agenda-item-disallowed-rules item)
			   :depth depth 
			   :assumptions (subgoal-assumptions s-item)
			   :difficulty 0) 
	      )
	  )
  )

(defun pr-subgoal (&optional brief &rest args)
  (print-subgoal 
   (cond ((or (member 'proof args) (member '*proof* args)
	      (member *proof* args)) *proof*)
	 ((or (member 'disproof args) (member '*disproof* args)
	      (member *disproof* args)) *disproof*)
	 (t nil))
   (not brief)))
(defun pr-proof (&optional brief &rest args)
  (declare (ignore args))
  (print-subgoal *proof* (not brief)))
(defun pr-disproof (&optional brief &rest args)
  (declare (ignore args))
  (print-subgoal *disproof* (not brief)))
(set-display-function 'subgoal 'pr-subgoal "Prints subgoals")
(set-display-function 'proof 'pr-proof "Prints proof subgoal")
(set-display-function 'disproof 'pr-disproof "Prints disproof subgoal")

(defun print-subgoal (&optional subgoal df ass wffs rules &aux
			      debug
			      (oi *indent*)
			      (os *spaces*)
			      assump wffu rulesu ss sn depth subgoal-wff)
	
  (setq wffs nil)
  (unless subgoal
	  (if *solutions*
	      (if (equal (caar *solutions*) 'NO) (setq subgoal *disproof*)
		  (setq subgoal *proof*))
	      ))
  (when subgoal
  (setq depth (subgoal-depth subgoal))
  (setq subgoal-wff (subgoal-subgoal subgoal))
  (if df (setq debug df)
      (setq debug (or (not *solutions*)
		      (and (equal (subgoal-proof-prop subgoal) 'PROOF)
			   (equal (caar *solutions*) 'NO))
		      (and (equal (subgoal-proof-prop subgoal) 'DISPROOF)
			   (not (equal (caar *solutions*) 'NO))))))
  )
  (when (and subgoal (if debug t (subgoal-useful subgoal)))
  (set-indent depth)
  (print-line "------------------------------")
  (if (equal (subgoal-subgoal subgoal) 'YES)
      (print-line (subgoal-proof-prop subgoal) " subgoal terminated with YES")
      (print-line (subgoal-proof-prop subgoal) " subgoal to prove: " 
		  (print-info subgoal-wff)))
  (print-line "Depth: " depth (if (subgoal-useful subgoal) 
					    " solved and helpful " " ")) 
  (print-line "Interest: " (subgoal-interest subgoal) "    Inherited Interest: " (subgoal-inherit-interest subgoal))
  (print-line "Probability: " (subgoal-c-prob subgoal) " " (subgoal-prob subgoal))
  (when (subgoal-subs subgoal) (print-it "      Subs: " (subgoal-subs subgoal)))
		
  (setq assump (if debug (subgoal-assumptions subgoal)
		   (set-difference (subgoal-assumptions subgoal) ass)))
  (when assump
	(print-line "Assumptions:")
	(dolist (a assump)
		(print-line "   " a ": " (getp a 'print)))
	)
  (when debug
	(setq rulesu 
	      (set-difference (subgoal-rules subgoal) rules :test 'equal))
	(when rulesu
	      (print-line "Rules:")
	      (mapcar #'(lambda (a) (print-line "   " (car a) ": " 
		(print-info (car a)) " " (print-info (cadr a)) " " (caddr a)))
		      rulesu)) )
  (setq wffu (if debug (subgoal-wffs subgoal)
		   (set-difference (subgoal-wffs subgoal) wffs)))
  (when wffu
	(print-line "Wff(s) used:")
	(mapcar #'(lambda (a) (print-line "   " a ": " (print-info a)
		  (if (getp a 'assumption) "     (assumption)" ""))) wffu))
  (when (and debug (subgoal-answer subgoal))
	(print-line "Answer: " (subgoal-answer subgoal)))
  (when (setq ss (subgoal-subgoals subgoal))
	(setq sn (length ss))
	(print-line "Subgoals - need to solve " sn " subgoals")
	(when (> sn 1) (print-it " - combine with " (subgoal-combo subgoal)))
	(mapcar #'(lambda (a) 
			  (unless (null a)
		 (print-subgoal a debug (subgoal-assumptions subgoal) 
				(subgoal-wffs subgoal) (subgoal-rules subgoal))))
		ss))
  (reset-indent os oi)
  nil))


;; combine-subgoal takes two wffs and combines them using the conjunctive
;; predicate how (or, and). If either wff already has how as its top level
;; predicate, simplifications are done accordingly.

(defun combine-subgoal (how wff1 wff2)
  ; if top level predicate is already how, just add other args
  (if wff1
      (if wff2
	  (if nil 
	      (combine-subgoal 'or (negate wff1) wff2)
	  ; if how is the same as one of the other predicates, don't
	  ; add extra levels
	  (let ((pred1 (if (atom wff1) nil (prop-pred wff1)))
		(pred2 (if (atom wff2) nil (prop-pred wff2)))
		(args1 (list wff1))
		(args2 (list wff2)))
	       (when (and (equal how 'implies) (equal pred2 'implies))
		     (setq args1 (list 
                         (combine-subgoal 'and wff1 (car (prop-args wff2))))) 
		     (setq args2 (cdr (prop-args wff2))) )
	       (unless (equal how 'implies)
		       (if (equal pred1 how) 
			   (setq args1 (prop-args wff1))) 
		       (if (equal pred2 how) 
			   (setq args2 (prop-args wff2))) )
	       (make-prop nil :pred how :args (append args1 args2))
	       ) )
	  wff1)
      wff2)
  )

(defun combine-subs (s1 s2 &optional both &aux news1)
  ; change anything in s1 whose cadr is a car in s2 to its cadr
  (dolist (s s1)
	  (setq news1 (append news1 (list (list (car s)
			(normalize-term (substitute-term (cadr s) s2 nil 1 nil)
					'+) (caddr s)))))
	  )
  (if both (setq news1 (append news1 
	  (remove-if-not #'(lambda (a) 
		   (member (car a) (append *wh-pattern* *e-pattern*))) s2))))
  news1)

(defun pick-out-evars (item)
  (cond ((symbolp item)
	 (case (getp item 'wff-type)
	       (quantified
		(if (member (getp item 'quantifier) '(E E! THE))
		    (append (pick-out-evars (getp item 'main))
			    (pick-out-evars (getp item 'restrict))
			    (list (getp item 'variable)))))
	       (logical
		(if (member (getp item 'pred) '(and or))
		    (apply 'append (mapcar 'pick-out-evars 
					   (getp item 'arguments)))))
	       ))
	))

(defun switch-quant (ante conse &aux evars)
  ; go through ante looking for vars quantified by E
  ; 
  (setq evars (pick-out-evars ante))
  (setq ante (substitute-prop ante (mapcar #'(lambda (a) (list a a)) evars)))
  (setq conse (getp conse 'print))
  (dolist (e evars)
	  (setq conse (list 'E e conse)))
  (cons conse (normalize ante))
  )
